package lcsb.mapviewer.api;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.validation.FieldError;
import org.springframework.validation.ObjectError;
import org.springframework.web.HttpMediaTypeNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.MissingServletRequestParameterException;
import org.springframework.web.bind.ServletRequestBindingException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.method.annotation.MethodArgumentTypeMismatchException;

import com.fasterxml.jackson.databind.JsonMappingException;
import com.google.gson.Gson;

import lcsb.mapviewer.annotation.services.dapi.DapiConnectionException;
import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.common.MimeType;
import lcsb.mapviewer.model.Stacktrace;
import lcsb.mapviewer.model.cache.FileEntry;
import lcsb.mapviewer.services.FailedDependencyException;
import lcsb.mapviewer.services.ObjectExistsException;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IStacktraceService;

public abstract class BaseController {
  protected static Logger logger = LogManager.getLogger();

  @Autowired
  private IStacktraceService stacktraceService;

  @ExceptionHandler({ Exception.class })
  public ResponseEntity<Object> handleException(final Exception e) {
    if (e instanceof AccessDeniedException) {
      return createErrorResponse("Access denied.", e.getMessage(), new HttpHeaders(), HttpStatus.FORBIDDEN);
    } else if (e instanceof ObjectNotFoundException) {
      return createErrorResponse("Object not found.", e.getMessage(), new HttpHeaders(), HttpStatus.NOT_FOUND);
    } else if (e instanceof ObjectExistsException) {
      return createErrorResponse("Object already exists.", e.getMessage(), new HttpHeaders(), HttpStatus.CONFLICT);
    } else if (e instanceof FailedDependencyException) {
      return createErrorResponse("Problem with external service.", e.getMessage(), new HttpHeaders(), HttpStatus.FAILED_DEPENDENCY);
    } else if (e instanceof UpdateConflictException) {
      return createErrorResponse("There is a conflict with another query. Try again later.", e.getMessage(),
          new HttpHeaders(), HttpStatus.CONFLICT);
    } else if (e instanceof OperationNotAllowedException) {
      return createErrorResponse("Operation not allowed.", e.getMessage(), new HttpHeaders(),
          HttpStatus.METHOD_NOT_ALLOWED);
    } else if (e instanceof DapiConnectionException) {
      logger.debug(e, e);
      return createErrorResponse("There is problem with DAPI connection.", e.getMessage(), new HttpHeaders(),
          HttpStatus.UNAVAILABLE_FOR_LEGAL_REASONS);
    } else if (e instanceof ConstraintViolationException) {
      StringBuilder sb = new StringBuilder("");
      for (ConstraintViolation<?> violation : ((ConstraintViolationException) e).getConstraintViolations()) {
        sb.append("[").append(violation.getMessage()).append("] ");
      }
      return createErrorResponse("Query server error.", sb.toString(), new HttpHeaders(), HttpStatus.BAD_REQUEST);
    } else if (e instanceof MethodArgumentNotValidException) {
      StringBuilder sb = new StringBuilder("");
      for (ObjectError error : ((MethodArgumentNotValidException) e).getBindingResult().getAllErrors()) {
        if (error instanceof FieldError) {
          sb.append("[").append(((FieldError) error).getField() + " - " + error.getDefaultMessage()).append("] ");
        } else {
          sb.append("[").append(error).append("] ");
        }
      }
      return createErrorResponse("Query server error.", sb.toString(), new HttpHeaders(), HttpStatus.BAD_REQUEST);
    } else if (e instanceof QueryException
        || e instanceof HttpMessageNotReadableException
        || e instanceof MissingServletRequestParameterException
        || e instanceof HttpMediaTypeNotSupportedException
        || e instanceof JsonMappingException
        || e instanceof MethodArgumentTypeMismatchException) {
      logger.debug(e, e);
      return createErrorResponse("Query server error.", e.getMessage(), new HttpHeaders(), HttpStatus.BAD_REQUEST);
    } else if (e instanceof ServletRequestBindingException && e.getMessage().contains(Configuration.AUTH_TOKEN)) {
      return createErrorResponse("Access denied.", e.getMessage(), new HttpHeaders(), HttpStatus.FORBIDDEN);
    } else {
      logger.error(e, e);
      return createErrorResponse("Internal server error.", e, new HttpHeaders(), HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  private ResponseEntity<Object> createErrorResponse(final String errorMessage, final Exception e, final HttpHeaders httpHeaders,
      final HttpStatus status) {
    Stacktrace st = stacktraceService.createFromStackTrace(e);

    Map<String, String> response = new HashMap<>();
    response.put("error", errorMessage);
    response.put("reason", e.getMessage());
    response.put("error-id", st.getId());
    return new ResponseEntity<Object>(new Gson().toJson(response), httpHeaders, status);
  }

  private ResponseEntity<Object> createErrorResponse(final String errorMessage, final String error, final HttpHeaders httpHeaders,
      final HttpStatus status) {

    Map<String, String> response = new HashMap<>();
    response.put("error", errorMessage);
    response.put("reason", error);
    return new ResponseEntity<Object>(new Gson().toJson(response), httpHeaders, status);
  }

  protected ResponseEntity<byte[]> sendAsZip(final FileEntry originalFile) throws IOException {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    ZipOutputStream zos = new ZipOutputStream(baos);

    ZipEntry entry = new ZipEntry(originalFile.getOriginalFileName());

    zos.putNextEntry(entry);
    zos.write(originalFile.getFileContent());
    zos.closeEntry();
    zos.close();

    return ResponseEntity.ok().contentLength(baos.size())
        .contentType(MediaType.parseMediaType(MimeType.ZIP.getTextRepresentation()))
        .header("Content-Disposition", "attachment; filename=" + originalFile.getOriginalFileName() + ".zip")
        .body(baos.toByteArray());
  }

  protected Map<String, Object> okStatus() {
    Map<String, Object> result = new HashMap<>();
    result.put("status", "OK");
    return result;
  }

  protected boolean isMimeTypeIncluded(final String accept, final String mimeType) {
    List<MediaType> mediaTypes = MediaType.parseMediaTypes(accept);
    for (final MediaType mediaType : mediaTypes) {
      if (Objects.equals(mediaType.getType() + "/" + mediaType.getSubtype(), mimeType)) {
        return true;
      }
    }
    return false;
  }

}
