package lcsb.mapviewer.services.impl;

import java.awt.Color;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.mail.MessagingException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.unboundid.ldap.sdk.LDAPException;

import lcsb.mapviewer.commands.ColorExtractor;
import lcsb.mapviewer.common.geometry.ColorParser;
import lcsb.mapviewer.model.map.layout.ProjectBackground;
import lcsb.mapviewer.model.security.Privilege;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.ConfigurationElementType;
import lcsb.mapviewer.model.user.ConfigurationElementTypeGroup;
import lcsb.mapviewer.model.user.ResetPasswordToken;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.model.user.UserClassAnnotators;
import lcsb.mapviewer.model.user.UserClassRequiredAnnotations;
import lcsb.mapviewer.model.user.UserClassValidAnnotations;
import lcsb.mapviewer.model.user.annotator.AnnotatorData;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.persist.dao.map.DataOverlayDao;
import lcsb.mapviewer.persist.dao.map.ProjectBackgroundDao;
import lcsb.mapviewer.persist.dao.user.ResetPasswordTokenDao;
import lcsb.mapviewer.persist.dao.user.UserDao;
import lcsb.mapviewer.services.InvalidTokenException;
import lcsb.mapviewer.services.interfaces.IConfigurationService;
import lcsb.mapviewer.services.interfaces.ILdapService;
import lcsb.mapviewer.services.interfaces.IPrivilegeService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.utils.EmailSender;

@Transactional
@Service
public class UserService implements IUserService {

  private static Logger logger = LogManager.getLogger();

  private UserDao userDao;
  private DataOverlayDao dataOverlayDao;
  private ILdapService ldapService;
  private IConfigurationService configurationService;
  private IPrivilegeService privilegeService;
  private ProjectDao projectDao;
  private ProjectBackgroundDao projectBackgroundDao;
  private ResetPasswordTokenDao resetPasswordTokenDao;
  private EmailSender emailSender;

  @Autowired
  public UserService(final UserDao userDao,
      final ILdapService ldapService,
      final IConfigurationService configurationService,
      final IPrivilegeService privilegeService,
      final ProjectDao projectDao,
      final ResetPasswordTokenDao resetPasswordTokenDao,
      final EmailSender emailSender,
      final DataOverlayDao dataOverlayDao,
      final ProjectBackgroundDao projectBackgroundDao) {
    this.userDao = userDao;
    this.ldapService = ldapService;
    this.configurationService = configurationService;
    this.privilegeService = privilegeService;
    this.projectDao = projectDao;
    this.resetPasswordTokenDao = resetPasswordTokenDao;
    this.emailSender = emailSender;
    this.dataOverlayDao = dataOverlayDao;
    this.projectBackgroundDao = projectBackgroundDao;
  }

  @Override
  public void addUser(final User user) {
    userDao.add(user);
    logger.info("User " + user.getLogin() + " created successfully");
  }

  @Override
  public void updateUser(final User user) {
    userDao.update(user);
  }

  @Override
  public void deleteUser(final User user) {
    String login = user.getLogin();
    User userFromDb = getUserByLogin(login);
    List<ProjectBackground> backgrounds = projectBackgroundDao.getProjectBackgroundsByUser(userFromDb);
    for (final ProjectBackground background : backgrounds) {
      projectBackgroundDao.delete(background);
    }
    dataOverlayDao.deleteDataOverlays(dataOverlayDao.getDataOverlayByOwner(userFromDb));
    userDao.delete(userFromDb);
    logger.info("User " + login + " removed successfully");
  }

  @Override
  public User getUserByLogin(final String login, final boolean loadLazy) {
    User user = getUserByLogin(login);
    if (loadLazy) {
      Hibernate.initialize(user.getPrivileges());
      if (user.getAnnotationSchema() != null) {
        for (final UserClassAnnotators annotator : user.getAnnotationSchema().getClassAnnotators()) {
          for (final AnnotatorData data : annotator.getAnnotators()) {
            Hibernate.initialize(data.getAnnotatorParams());
          }
        }
        for (final UserClassValidAnnotations annotation : user.getAnnotationSchema().getClassValidAnnotators()) {
          Hibernate.initialize(annotation.getValidMiriamTypes());
        }
        for (final UserClassRequiredAnnotations annotation : user.getAnnotationSchema().getClassRequiredAnnotators()) {
          Hibernate.initialize(annotation.getRequiredMiriamTypes());
        }
        Hibernate.initialize(user.getAnnotationSchema().getGuiPreferences());
      }
    }
    return user;
  }

  @Override
  public User getUserByLogin(final String login) {
    return userDao.getUserByLogin(login);
  }

  @Override
  public List<User> getUsers(final boolean loadLazy) {
    List<User> users = userDao.getAll();
    if (loadLazy) {
      for (final User user : users) {
        Hibernate.initialize(user.getPrivileges());
      }
    }
    return users;
  }

  @Override
  public void grantUserPrivilege(final User user, final PrivilegeType type) {
    User dbUser = getUserByLogin(user.getLogin());
    dbUser.addPrivilege(privilegeService.getPrivilege(type));
    userDao.update(dbUser);
  }

  @Override
  public void grantUserPrivilege(final User user, final PrivilegeType type, final String objectId) {
    User dbUser = getUserByLogin(user.getLogin());
    dbUser.addPrivilege(privilegeService.getPrivilege(type, objectId));
    userDao.update(dbUser);
  }

  @Override
  public void revokeUserPrivilege(final User user, final PrivilegeType type) {
    User dbUser = getUserByLogin(user.getLogin());
    dbUser.removePrivilege(privilegeService.getPrivilege(type));
    userDao.update(dbUser);
  }

  @Override
  public void revokeUserPrivilege(final User user, final PrivilegeType type, final String objectId) {
    User dbUser = getUserByLogin(user.getLogin());
    dbUser.removePrivilege(privilegeService.getPrivilege(type, objectId));
    userDao.update(dbUser);
  }

  @Override
  public void grantDefaultPrivileges(final User user) {
    Arrays.stream(ConfigurationElementType.values())
        .filter(type -> type.getGroup() == ConfigurationElementTypeGroup.DEFAULT_USER_PRIVILEGES)
        .filter(type -> Boolean.valueOf(configurationService.getConfigurationValue(type)))
        .forEach(type -> grantPrivilegesForDefaultConfigurationElementType(user, type));
  }

  @Override
  public void grantPrivilegeToAllUsersWithDefaultAccess(final PrivilegeType type, final String objectId) {
    getUsers(false).stream()
        .filter(u -> u.getPrivileges().contains(new Privilege(type, "*")))
        .forEach(u -> grantUserPrivilege(u, type, objectId));
  }

  @Override
  public void revokeObjectDomainPrivilegesForAllUsers(final PrivilegeType privilegeType, final String objectId) {
    userDao.getAll().forEach(user -> revokeUserPrivilege(user, privilegeType, objectId));
  }

  @Override
  public ColorExtractor getColorExtractorForUser(final User loggedUser) {
    Color colorMin = null;
    Color colorMax = null;
    Color colorSimple = null;
    Color colorNeutral = null;
    if (loggedUser != null) {
      User dbUser = getUserByLogin(loggedUser.getLogin());
      if (dbUser != null) {
        colorMin = dbUser.getMinColor();
        colorMax = dbUser.getMaxColor();
        colorSimple = dbUser.getSimpleColor();
        colorNeutral = dbUser.getNeutralColor();
      }
    }
    ColorParser parser = new ColorParser();

    if (colorMin == null) {
      colorMin = parser.parse(configurationService.getConfigurationValue(ConfigurationElementType.MIN_COLOR_VAL));
    }
    if (colorMax == null) {
      colorMax = parser.parse(configurationService.getConfigurationValue(ConfigurationElementType.MAX_COLOR_VAL));
    }
    if (colorSimple == null) {
      colorSimple = parser.parse(configurationService.getConfigurationValue(ConfigurationElementType.SIMPLE_COLOR_VAL));
    }
    if (colorNeutral == null) {
      colorNeutral = parser
          .parse(configurationService.getConfigurationValue(ConfigurationElementType.NEUTRAL_COLOR_VAL));
    }
    return new ColorExtractor(colorMin, colorMax, colorSimple, colorNeutral);
  }

  @Override
  public Map<String, Boolean> ldapAccountExistsForLogin(final Collection<User> logins) {
    Map<String, Boolean> result = new HashMap<>();
    for (final User user : logins) {
      result.put(user.getLogin(), false);
    }

    if (ldapService.isValidConfiguration()) {
      try {
        List<String> ldapUserNames = ldapService.getUsernames();
        for (final String string : ldapUserNames) {
          String ldapLogin = string.toLowerCase();
          if (result.keySet().contains(ldapLogin)) {
            result.put(ldapLogin, true);
          }
        }
      } catch (final LDAPException e) {
        logger.error("Problem with accessing LDAP directory", e);
      }
    }
    return result;
  }

  private void grantPrivilegesForDefaultConfigurationElementType(final User user, final ConfigurationElementType type) {
    switch (type) {
      case DEFAULT_CAN_CREATE_OVERLAYS:
        grantUserPrivilege(user, PrivilegeType.CAN_CREATE_OVERLAYS);
        break;
      case DEFAULT_READ_PROJECT:
        grantUserPrivilege(user, PrivilegeType.READ_PROJECT, "*");
        projectDao.getAll()
            .forEach(project -> grantUserPrivilege(user, PrivilegeType.READ_PROJECT, project.getProjectId()));
        break;
      case DEFAULT_WRITE_PROJECT:
        grantUserPrivilege(user, PrivilegeType.WRITE_PROJECT, "*");
        projectDao.getAll()
            .forEach(project -> grantUserPrivilege(user, PrivilegeType.WRITE_PROJECT, project.getProjectId()));
        break;
      default:
        throw new IllegalArgumentException("The configuration element type '+" + type + "' is not handled.");
    }
  }

  @Override
  public void createResetPasswordToken(final User user) throws MessagingException {
    Calendar expires = Calendar.getInstance();
    expires.add(Calendar.DAY_OF_MONTH, 1);
    ResetPasswordToken token = new ResetPasswordToken(user, UUID.randomUUID().toString(), expires);
    resetPasswordTokenDao.add(token);

    String content = "Dear " + user.getName() + " " + user.getSurname() + " (" + user.getLogin() + "),<br/><br/>"
        + "We received request to reset your password. You can reset the password using this <a href=\""
        + configurationService.getConfigurationValue(ConfigurationElementType.MINERVA_ROOT)
        + "login.xhtml?resetPasswordToken=" + token.getToken()
        + "\">link</a>. Link will be valid for the next 24 hours.<br/><br/>"
        + "If you did not perform this action you can ignore the email.<br/>";
    try {
      emailSender.sendEmail("Minerva - Reset password", content, user.getEmail());
    } catch (final MessagingException e) {
      resetPasswordTokenDao.delete(token);
      throw e;
    }

  }

  @Override
  public void resetPassword(final String token, final String cryptedPassword) throws InvalidTokenException {
    ResetPasswordToken resetToken = resetPasswordTokenDao.getByToken(token);
    if (resetToken != null && resetToken.getExpires().before(Calendar.getInstance())) {
      resetPasswordTokenDao.delete(resetToken);
      resetToken = null;
    }
    if (resetToken == null) {
      throw new InvalidTokenException("Invalid token");
    }
    resetToken.getUser().setCryptedPassword(cryptedPassword);
    userDao.update(resetToken.getUser());
    resetPasswordTokenDao.delete(resetToken);
  }

  @Override
  public long getPasswordTokenCount() {
    return resetPasswordTokenDao.getCount();
  }

  @Override
  public void add(final ResetPasswordToken token) {
    resetPasswordTokenDao.add(token);
  }

  @Override
  public List<ResetPasswordToken> getPasswordTokens(final String login) {
    return resetPasswordTokenDao.getByUser(getUserByLogin(login));
  }

}
