package lcsb.mapviewer.services.impl;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.CALLS_REAL_METHODS;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.core.impl.Log4jLogEvent;
import org.apache.logging.log4j.message.SimpleMessage;
import org.apache.poi.util.IOUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.primefaces.model.TreeNode;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.util.ReflectionTestUtils;

import lcsb.mapviewer.annotation.services.ModelAnnotator;
import lcsb.mapviewer.commands.CopyCommand;
import lcsb.mapviewer.common.MinervaLoggerAppender;
import lcsb.mapviewer.converter.ComplexZipConverter;
import lcsb.mapviewer.converter.ComplexZipConverterParams;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.converter.zip.ModelZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFileFactory;
import lcsb.mapviewer.model.LogMarker;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.ProjectLogEntry;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.ProjectStatus;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.OverviewImage;
import lcsb.mapviewer.model.map.OverviewImageLink;
import lcsb.mapviewer.model.map.OverviewLink;
import lcsb.mapviewer.model.map.layout.ProjectBackground;
import lcsb.mapviewer.model.map.layout.ProjectBackgroundImageLayer;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.model.user.UserAnnotationSchema;
import lcsb.mapviewer.model.user.UserClassAnnotators;
import lcsb.mapviewer.model.user.UserClassValidAnnotations;
import lcsb.mapviewer.model.user.annotator.AnnotatorData;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.persist.dao.ProjectLogEntryDao;
import lcsb.mapviewer.persist.dao.map.ModelDao;
import lcsb.mapviewer.persist.dao.map.ProjectBackgroundDao;
import lcsb.mapviewer.services.ServiceTestFunctions;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.overlay.AnnotatedObjectTreeRow;
import lcsb.mapviewer.services.utils.CreateProjectParams;
import lcsb.mapviewer.services.utils.data.BuildInBackgrounds;

@Rollback(true)
public class ProjectServiceTest extends ServiceTestFunctions {

  private final String tmpResultDir = "tmp/";

  @Autowired
  private ProjectBackgroundDao backgroundDao;

  private ZipEntryFileFactory zefFactory = new ZipEntryFileFactory();

  @Autowired
  private ModelAnnotator modelAnnotator;

  @Autowired
  protected IModelService modelService;

  @Autowired
  protected IUserService userService;

  @Autowired
  protected ModelDao modelDao;

  @Before
  public void setUp() throws Exception {
  }

  @After
  public void tearDown() throws Exception {
  }

  @Test
  public void testGetAllProjects() {
    List<Project> projects = projectService.getAllProjects();

    Project project = new Project();
    project.setProjectId(TEST_PROJECT_ID);
    project.setOwner(userDao.getUserByLogin(ADMIN_BUILT_IN_LOGIN));
    projectDao.add(project);
    projectDao.evict(project);

    List<Project> projects2 = projectService.getAllProjects();

    assertEquals(projects.size() + 1, projects2.size());

    for (final Project tmpProject : projects2) {
      if (!projects.contains(tmpProject)) {
        assertEquals(TEST_PROJECT_ID, tmpProject.getProjectId());
      }
    }

    Project project2 = projectDao.getProjectByProjectId(TEST_PROJECT_ID);
    assertNotNull(project2);
    assertNotEquals(project2, project);
    assertEquals(project.getId(), project2.getId());

    projectDao.delete(project2);
  }

  @Test
  public void testUpdater() throws Exception {
    projectService
        .createProject(createProjectParams("testFiles/centeredAnchorInModifier.xml").analyzeAnnotations(true));
    Project project = projectService.getProjectByProjectId(TEST_PROJECT_ID);
    assertNotNull(project);
    assertEquals(ProjectStatus.DONE, project.getStatus());
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testCreateComplex() throws Exception {
    try {
      ZipEntryFile entry1 = new ModelZipEntryFile("main.xml", "main", true, false, SubmodelType.UNKNOWN);
      ZipEntryFile entry2 = new ModelZipEntryFile("s1.xml", "s1", false, false, SubmodelType.UNKNOWN);
      ZipEntryFile entry3 = new ModelZipEntryFile("s2.xml", "s2", false, false, SubmodelType.UNKNOWN);
      ZipEntryFile entry4 = new ModelZipEntryFile("s3.xml", "s3", false, false, SubmodelType.UNKNOWN);
      ZipEntryFile entry5 = new ModelZipEntryFile("mapping.xml", "mapping", false, true, SubmodelType.UNKNOWN);
      projectService.createProject(createProjectParams("testFiles/complexModel/complex_model.zip").addZipEntry(entry1)
          .addZipEntry(entry2).addZipEntry(entry3).addZipEntry(entry4).addZipEntry(entry5)
          .semanticZoomContainsMultipleBackgrounds(true).images(true));
      Project project = projectService.getProjectByProjectId(TEST_PROJECT_ID);

      Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
      assertNotNull(model);
      assertEquals("main", model.getName());
      assertEquals(3, model.getSubmodelConnections().size());

      Model s1Model = null;
      Model s2Model = null;
      Model s3Model = null;

      for (final ModelSubmodelConnection submodel : model.getSubmodelConnections()) {
        if (submodel.getName().equals("s1")) {
          s1Model = submodel.getSubmodel().getModel();
        }
        if (submodel.getName().equals("s2")) {
          s2Model = submodel.getSubmodel().getModel();
        }
        if (submodel.getName().equals("s3")) {
          s3Model = submodel.getSubmodel().getModel();
        }
      }
      assertNotNull(s1Model);
      assertNotNull(s2Model);
      assertNotNull(s3Model);

      Element al1 = model.getElementByElementId("sa1");
      assertNotNull(al1.getSubmodel());
      assertEquals(SubmodelType.DOWNSTREAM_TARGETS, al1.getSubmodel().getType());
      assertEquals(s1Model, al1.getSubmodel().getSubmodel().getModel());

      Element al2 = model.getElementByElementId("sa2");
      assertNull(al2.getSubmodel());

      Element al4 = model.getElementByElementId("sa4");
      assertNotNull(al4.getSubmodel());
      assertEquals(SubmodelType.PATHWAY, al4.getSubmodel().getType());
      assertEquals(s2Model, al4.getSubmodel().getSubmodel().getModel());

      Element submodel1Element1 = s1Model.getElementByElementId("sa1");
      assertNotNull(submodel1Element1.getSubmodel());
      assertEquals(SubmodelType.DOWNSTREAM_TARGETS, submodel1Element1.getSubmodel().getType());
      assertEquals(s3Model, submodel1Element1.getSubmodel().getSubmodel().getModel());

      assertEquals(ProjectStatus.DONE, project.getStatus());

      Collection<ProjectBackground> backgrounds = backgroundDao.getProjectBackgroundsByProject(project);

      // now check layouts
      for (final ProjectBackground l : backgrounds) {
        assertEquals(model.getSubmodels().size() + 1, l.getProjectBackgroundImageLayer().size());
      }

      // and now check layouts (check if every submodel have them, and point
      // into different directory)

      int semanticOverlays = model.getZoomLevels() + 1;
      int overlays = BuildInBackgrounds.values().length + semanticOverlays;
      assertEquals(overlays, backgrounds.size());
      Set<String> directories = new HashSet<>();
      for (final ProjectBackgroundImageLayer imageLayer : backgrounds.iterator().next().getProjectBackgroundImageLayer()) {
        directories.add(imageLayer.getDirectory());
      }
      assertEquals(1 + model.getSubmodels().size(), directories.size());
    } finally {
      Project project = projectService.getProjectByProjectId(TEST_PROJECT_ID);
      if (project != null) {
        projectService.removeProject(project, null, false);
      }
    }
  }

  @Test
  public void testCreateComplexWithLayoutOrder() throws Exception {
    String filename = "testFiles/complexModel/complex_model_with_images.zip";
    CreateProjectParams params = createProjectParams(filename);

    ZipFile zipFile = new ZipFile(filename);
    Enumeration<? extends ZipEntry> entries = zipFile.entries();
    while (entries.hasMoreElements()) {
      ZipEntry entry = entries.nextElement();
      params.addZipEntry(zefFactory.createZipEntryFile(entry, zipFile));
    }
    zipFile.close();

    projectService.createProject(params.networkBackgroundAsDefault(true));
    Project project = projectService.getProjectByProjectId(TEST_PROJECT_ID);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    for (OverviewImage oi : project.getOverviewImages()) {
      for (OverviewLink ol : oi.getLinks()) {
        if (ol instanceof OverviewImageLink) {
          logger.debug(((OverviewImageLink) ol).getLinkedOverviewImage());
        }
      }
    }
    assertNotNull(model);
    assertEquals("main", model.getName());
    assertEquals(ProjectStatus.DONE, project.getStatus());
    assertEquals("Cannot find overview images the model", 2, project.getOverviewImages().size());
    assertEquals("Number of layouts doesn't match", BuildInBackgrounds.values().length,
        project.getProjectBackgrounds().size());
    assertEquals(BuildInBackgrounds.NESTED.getName(), project.getProjectBackgrounds().get(1).getName());
    assertEquals(BuildInBackgrounds.NORMAL.getName(), project.getProjectBackgrounds().get(0).getName());
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testGenerateImages() throws Exception {
    File f = createTempDirectory();
    String path = f.getAbsolutePath() + "/" + tmpResultDir;
    try {
      Model model = getModelForFile("testFiles/sample.xml", false);
      Project project = new Project();
      project.addModel(model);

      ProjectBackground background = new ProjectBackground(BuildInBackgrounds.NORMAL.getName());
      background.addProjectBackgroundImageLayer(new ProjectBackgroundImageLayer(model, path + "/normal"));
      project.addProjectBackground(background);
      background = new ProjectBackground(BuildInBackgrounds.NESTED.getName());
      background.addProjectBackgroundImageLayer(new ProjectBackgroundImageLayer(model, path + "/nested"));
      project.addProjectBackground(background);

      CreateProjectParams params = createProjectParams("testFiles/sample.xml").images(true).projectDir(tmpResultDir);

      ProjectService projectService = new ProjectService(Mockito.mock(ProjectDao.class),
          Mockito.mock(ProjectLogEntryDao.class),
          null, null, null, null, null, null, null, null, null, null, null, null, null);
      ReflectionTestUtils.setField(projectService, "self", projectService);

      projectService.createImages(project, params);

      File tmp = new File(path + "/normal");
      assertTrue(tmp.exists());

      tmp = new File(path + "/nested");
      assertTrue(tmp.exists());
    } finally {
      f.delete();
    }

  }

  @Test
  public void testGenerateImagesInSubmodels() throws Exception {
    File f = createTempDirectory();
    String path = f.getAbsolutePath() + "/" + tmpResultDir;
    try {
      Model model = getModelForFile("testFiles/sample.xml", false);

      Model model2 = getModelForFile("testFiles/sample.xml", false);

      Project project = new Project();
      project.addModel(model);
      project.addModel(model2);

      ProjectBackground background = new ProjectBackground(BuildInBackgrounds.NORMAL.getName());
      background.addProjectBackgroundImageLayer(new ProjectBackgroundImageLayer(model, path + "/normal_A"));
      project.addProjectBackground(background);
      background = new ProjectBackground(BuildInBackgrounds.NESTED.getName());
      background.addProjectBackgroundImageLayer(new ProjectBackgroundImageLayer(model2, path + "/nested_B"));
      project.addProjectBackground(background);

      model.addSubmodelConnection(new ModelSubmodelConnection(model2, SubmodelType.UNKNOWN, "name"));

      CreateProjectParams params = createProjectParams("testFiles/sample.xml").images(true).projectDir(tmpResultDir);

      ProjectService projectService = new ProjectService(Mockito.mock(ProjectDao.class),
          Mockito.mock(ProjectLogEntryDao.class),
          null, null, null, null, null, null, null, null, null, null, null, null, null);
      ReflectionTestUtils.setField(projectService, "self", projectService);

      projectService.createImages(project, params);

      File tmp = new File(path + "/normal_A");
      assertTrue(tmp.exists());

      tmp = new File(path + "/nested_B");
      assertTrue(tmp.exists());
    } finally {
      f.delete();
    }

  }

  @Test
  public void testCopyComplexWithCompartments() throws Exception {
    ZipEntryFile entry1 = new ModelZipEntryFile("main.xml", "main", true, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry2 = new ModelZipEntryFile("s1.xml", "s1", false, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry3 = new ModelZipEntryFile("s2.xml", "s2", false, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry4 = new ModelZipEntryFile("s3.xml", "s3", false, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry5 = new ModelZipEntryFile("mapping.xml", "mapping", false, true, SubmodelType.UNKNOWN);

    CreateProjectParams params = createProjectParams("testFiles/complexModel/complex_model_with_compartment.zip")
        .addZipEntry(entry1).addZipEntry(entry2).addZipEntry(entry3).addZipEntry(entry4).addZipEntry(entry5)
        .analyzeAnnotations(false);

    ComplexZipConverter parser = new ComplexZipConverter(CellDesignerXmlParser.class);
    ComplexZipConverterParams complexParams = new ComplexZipConverterParams().zipFile(params.getProjectFile());
    for (final ZipEntryFile entry : params.getZipEntries()) {
      complexParams.entry(entry);
    }
    Model model = parser.createModel(complexParams);

    assertNotNull(model);
    assertEquals("main", model.getName());
    assertEquals(3, model.getSubmodelConnections().size());

    int oldSize = 0;
    for (final Model m : model.getSubmodels()) {
      oldSize += m.getParentModels().size();
    }

    new CopyCommand(model).execute();

    int newSize = 0;
    for (final Model m : model.getSubmodels()) {
      newSize += m.getParentModels().size();
    }
    assertEquals("Submodels doesn't match after copying", oldSize, newSize);
  }

  @Test
  public void testAnnotatorTree() throws Exception {
    TreeNode tree = projectService.createClassAnnotatorTree(null);
    assertNotNull(tree);

    AnnotatedObjectTreeRow dataNode = (AnnotatedObjectTreeRow) tree.getData();

    assertEquals(BioEntity.class, dataNode.getClazz());

    Queue<TreeNode> nodes = new LinkedList<TreeNode>();
    nodes.add(tree);
    boolean annotatorsFound = false;
    while (!nodes.isEmpty()) {
      TreeNode node = nodes.poll();
      dataNode = (AnnotatedObjectTreeRow) node.getData();
      for (AnnotatorData data : dataNode.getUsedAnnotators()) {
        assertFalse(data.getAnnotatorClassName().getName().contains("Proxy"));
      }
      if (dataNode.getValidAnnotators().size() > 0) {
        annotatorsFound = true;
      }
      for (final TreeNode n : node.getChildren()) {
        nodes.add(n);
      }
    }
    assertTrue(annotatorsFound);
  }

  @Test
  public void testEmptyAnnotatorTreeForUser() throws Exception {
    User user = createUser();
    UserAnnotationSchema schema = modelAnnotator.createDefaultAnnotatorSchema();
    for (final UserClassAnnotators classAnnotators : schema.getClassAnnotators()) {
      classAnnotators.setAnnotators(new ArrayList<>());
    }
    user.setAnnotationSchema(schema);
    userDao.update(user);

    TreeNode tree = projectService.createClassAnnotatorTree(user);
    assertNotNull(tree);

    AnnotatedObjectTreeRow dataNode = (AnnotatedObjectTreeRow) tree.getData();

    assertEquals(BioEntity.class, dataNode.getClazz());

    Queue<TreeNode> nodes = new LinkedList<TreeNode>();
    nodes.add(tree);
    boolean annotatorsFound = false;
    while (!nodes.isEmpty()) {
      TreeNode node = nodes.poll();
      dataNode = (AnnotatedObjectTreeRow) node.getData();
      assertEquals("Annotators found for class: " + dataNode.getClazz(), 0, dataNode.getUsedAnnotators().size());
      if (dataNode.getValidAnnotators().size() > 0) {
        annotatorsFound = true;
      }
      for (final TreeNode n : node.getChildren()) {
        nodes.add(n);
      }
    }
    assertTrue(annotatorsFound);

    userDao.delete(user);
  }

  @Test
  public void testAnnotatorTreeForUser() throws Exception {
    User user = createUser();
    UserAnnotationSchema schema = new UserAnnotationSchema();
    schema.addClassAnnotator(
        new UserClassAnnotators(Protein.class, modelAnnotator.getDefaultAnnotators(Protein.class)));
    user.setAnnotationSchema(schema);
    userDao.update(user);

    TreeNode tree = projectService.createClassAnnotatorTree(user);
    assertNotNull(tree);

    AnnotatedObjectTreeRow dataNode = (AnnotatedObjectTreeRow) tree.getData();

    assertEquals(BioEntity.class, dataNode.getClazz());

    Queue<TreeNode> nodes = new LinkedList<>();
    nodes.add(tree);
    boolean annotatorsFound = false;
    while (!nodes.isEmpty()) {
      TreeNode node = nodes.poll();
      dataNode = (AnnotatedObjectTreeRow) node.getData();
      if (dataNode.getUsedAnnotators().size() > 0) {
        annotatorsFound = true;
      }
      for (final TreeNode n : node.getChildren()) {
        nodes.add(n);
      }
    }
    assertTrue(annotatorsFound);

    userDao.delete(user);
  }

  @Test
  public void testEmptyValidMiriamTreeForUser() throws Exception {
    User user = createUser();
    UserAnnotationSchema schema = new UserAnnotationSchema();
    user.setAnnotationSchema(schema);
    userDao.update(user);

    TreeNode tree = projectService.createClassAnnotatorTree(user);
    assertNotNull(tree);

    AnnotatedObjectTreeRow dataNode = (AnnotatedObjectTreeRow) tree.getData();

    assertEquals(BioEntity.class, dataNode.getClazz());

    Queue<TreeNode> nodes = new LinkedList<TreeNode>();
    nodes.add(tree);
    boolean annotatorsFound = false;
    while (!nodes.isEmpty()) {
      TreeNode node = nodes.poll();
      dataNode = (AnnotatedObjectTreeRow) node.getData();
      assertEquals("Valid miriam found for class: " + dataNode.getClazz(), 0, dataNode.getValidAnnotations().size());
      assertEquals("Required miriam found for class: " + dataNode.getClazz(), 0,
          dataNode.getRequiredAnnotations().size());
      if (dataNode.getMissingValidAnnotations().size() > 0 && dataNode.getMissingRequiredAnnotations().size() > 0) {
        annotatorsFound = true;
      }
      for (final TreeNode n : node.getChildren()) {
        nodes.add(n);
      }
    }
    assertTrue(annotatorsFound);

    userDao.delete(user);
  }

  @Test
  public void testValidMiriamTreeForUser() throws Exception {
    User user = createUser();
    UserAnnotationSchema schema = new UserAnnotationSchema();
    schema.addClassValidAnnotations(new UserClassValidAnnotations(Protein.class, MiriamType.values()));
    user.setAnnotationSchema(schema);
    userDao.update(user);
    userDao.flush();

    TreeNode tree = projectService.createClassAnnotatorTree(user);
    assertNotNull(tree);

    AnnotatedObjectTreeRow dataNode = (AnnotatedObjectTreeRow) tree.getData();

    assertEquals(BioEntity.class, dataNode.getClazz());

    Queue<TreeNode> nodes = new LinkedList<>();
    nodes.add(tree);
    boolean annotatorsFound = false;
    while (!nodes.isEmpty()) {
      TreeNode node = nodes.poll();
      dataNode = (AnnotatedObjectTreeRow) node.getData();
      if (dataNode.getValidAnnotations().size() > 0) {
        annotatorsFound = true;
      }
      for (final TreeNode n : node.getChildren()) {
        nodes.add(n);
      }
    }
    assertTrue(annotatorsFound);

    userDao.delete(user);
  }

  protected Project createComplexProject(final String projectId, final String filename) throws IOException, SecurityException {
    CreateProjectParams params = createProjectParams(filename);

    ZipFile zipFile = new ZipFile(filename);
    Enumeration<? extends ZipEntry> entries = zipFile.entries();
    while (entries.hasMoreElements()) {
      ZipEntry entry = entries.nextElement();
      ZipEntryFile e = zefFactory.createZipEntryFile(entry, zipFile);
      params.addZipEntry(e);
    }
    zipFile.close();

    projectService.createProject(params);
    Project project = projectService.getProjectByProjectId(projectId);
    return project;
  }

  @Test
  public void testCreateComplexWithImages() throws Exception {
    String filename = "testFiles/complexModel/complex_model_with_images.zip";
    Project project = createComplexProject(TEST_PROJECT_ID, filename);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    assertNotNull(model);
    assertEquals("main", model.getName());
    assertEquals(ProjectStatus.DONE, project.getStatus());
    assertEquals("Cannot find overview images the model", 2, project.getOverviewImages().size());
    assertEquals("Number of layouts doesn't match", BuildInBackgrounds.values().length,
        project.getProjectBackgrounds().size());
    assertEquals(BuildInBackgrounds.NESTED.getName(), project.getProjectBackgrounds().get(0).getName());
    assertEquals(BuildInBackgrounds.NORMAL.getName(), project.getProjectBackgrounds().get(1).getName());
    projectService.removeProject(project, null, false);
  }

  private CreateProjectParams createProjectParams(final String filename) throws FileNotFoundException, IOException {
    InputStream is = new FileInputStream(filename);
    return createProjectParams(filename, is);
  }

  private CreateProjectParams createProjectParams(final String filename, final InputStream is) throws IOException {
    UploadedFileEntry fileEntry = new UploadedFileEntry();
    fileEntry.setLocalPath(filename);
    byte[] data = IOUtils.toByteArray(is);
    fileEntry.setLength(data.length);
    fileEntry.setFileContent(data);

    return new CreateProjectParams()
        // .setUser(createUser("admin"))
        .setUser(userService.getUserByLogin(ADMIN_BUILT_IN_LOGIN))
        .projectId(TEST_PROJECT_ID)
        .version("")
        .complex(filename.endsWith("zip"))
        .projectFile(fileEntry).annotations(true).images(false).async(false).projectDir(tmpResultDir)
        .analyzeAnnotations(true).parser(new CellDesignerXmlParser());
  }

  @Test(timeout = 30000)
  public void testCreateComplexWithLayouts() throws Exception {
    String filename = "testFiles/complexModel/complex_model_with_layouts.zip";
    int overlaysInZipFile = 1;
    Project project = createComplexProject(TEST_PROJECT_ID, filename);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    assertNotNull(model);
    assertEquals("main", model.getName());
    assertEquals(ProjectStatus.DONE, project.getStatus());
    assertEquals("Number of backgrounds doesn't match", BuildInBackgrounds.values().length,
        project.getProjectBackgrounds().size());
    assertEquals("Number of data overlays doesn't match", overlaysInZipFile,
        project.getDataOverlays().size());
    boolean additionalFound = false;

    for (final DataOverlay layout : project.getDataOverlays()) {
      if ("example name".equals(layout.getName())) {
        additionalFound = true;
        assertEquals("layout description", layout.getDescription());
      }
    }
    assertTrue("Cannot find layout with predefined name from file", additionalFound);
    projectService.removeProject(project, null, false);
    assertNull(projectService.getProjectByProjectId(TEST_PROJECT_ID));
  }

  @Test
  public void testCreateComplexWithSumbaps() throws Exception {
    String filename = "testFiles/complexModel/complex_model_with_submaps.zip";
    Project project = createComplexProject(TEST_PROJECT_ID, filename);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    assertNotNull(model);
    assertEquals("main", model.getName());
    assertEquals(ProjectStatus.DONE, project.getStatus());
    assertEquals("Number of layouts doesn't match", BuildInBackgrounds.values().length,
        project.getProjectBackgrounds().size());
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testCreateEmptyComplex() throws Exception {
    String filename = "testFiles/complexModel/empty_complex_model.zip";
    Project project = createComplexProject(TEST_PROJECT_ID, filename);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    assertNotNull(model);
    assertEquals(ProjectStatus.DONE, project.getStatus());
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testCreateEmptyComplex2() throws Exception {
    String filename = "testFiles/complexModel/empty_complex_model2.zip";
    Project project = createComplexProject(TEST_PROJECT_ID, filename);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    assertNotNull(model);
    assertEquals(ProjectStatus.DONE, project.getStatus());
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testCreateEmptyComplex3() throws Exception {
    String filename = "testFiles/complexModel/empty_complex_model3.zip";
    Project project = createComplexProject(TEST_PROJECT_ID, filename);

    Model model = new ModelFullIndexed(modelDao.getLastModelForProjectIdentifier(TEST_PROJECT_ID, false));
    assertNotNull(model);
    assertEquals(ProjectStatus.DONE, project.getStatus());
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testUpdateProjectWithoutModel() throws Exception {
    createUser();

    Project project = new Project();
    project.setProjectId(TEST_PROJECT_ID);
    project.setOwner(userDao.getUserByLogin(ADMIN_BUILT_IN_LOGIN));
    projectDao.add(project);
    projectService.updateProject(project);
    projectService.removeProject(project, null, false);
  }

  @Test
  public void testCreateLogEntries() throws Exception {
    Protein protein = new GenericProtein("id");
    MinervaLoggerAppender appender = MinervaLoggerAppender.createAppender();
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.WARN)
        .setMessage(new SimpleMessage("Hello world"))
        .setMarker(new LogMarker(ProjectLogEntryType.OTHER, protein))
        .build());
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.ERROR)
        .setMessage(new SimpleMessage("Hello world 2"))
        .setMarker(new LogMarker(ProjectLogEntryType.OTHER, protein))
        .build());
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.ERROR)
        .setMessage(new SimpleMessage("Hello world 3"))
        .build());

    ProjectService service = Mockito.mock(ProjectService.class, Mockito.CALLS_REAL_METHODS);
    Set<ProjectLogEntry> entries = service.createLogEntries(appender);
    assertEquals(3, entries.size());
  }

  @Test
  public void testCreateIgnoredLogEntries() throws Exception {
    MinervaLoggerAppender appender = MinervaLoggerAppender.createAppender();
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.WARN)
        .setMessage(new SimpleMessage("Hello world"))
        .setMarker(LogMarker.IGNORE)
        .build());
    ProjectService service = Mockito.mock(ProjectService.class, Mockito.CALLS_REAL_METHODS);
    Set<ProjectLogEntry> entries = service.createLogEntries(appender);
    assertEquals(0, entries.size());
  }

  @Test
  public void testFixProjectIssuesWithElementWithoutModel() throws Exception {
    ProjectService service = Mockito.mock(ProjectService.class, CALLS_REAL_METHODS);
    Project project = new Project();
    ModelData model = new ModelData();
    GenericProtein protein = new GenericProtein("1");
    project.addModel(model);
    model.addElement(protein);
    service.fixProjectIssues(project);
    assertTrue(project.getLogEntries().size() > 0);
  }

}
