package lcsb.mapviewer.web;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.delete;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.patch;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.post;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.requestFields;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.requestParameters;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.JsonFieldType;
import org.springframework.restdocs.payload.ResponseFieldsSnippet;
import org.springframework.restdocs.request.PathParametersSnippet;
import org.springframework.restdocs.request.RequestParametersSnippet;
import org.springframework.restdocs.snippet.Snippet;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import com.fasterxml.jackson.core.type.TypeReference;
import com.google.gson.Gson;
import com.google.gson.JsonParser;

import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.layout.ReferenceGenomeType;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.overlay.DataOverlayType;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.interfaces.IDataOverlayService;
import lcsb.mapviewer.services.interfaces.IUserService;

@RunWith(SpringJUnit4ClassRunner.class)
public class OverlayControllerIntegrationTest extends ControllerIntegrationTest {

  private static final String TEST_USER_PASSWORD = "test_user_pass";
  private static final String TEST_USER_LOGIN = "test_user";
  private static final String TEST_CURATOR_PASSWORD = "test_curator_pass";
  private static final String TEST_CURATOR_LOGIN = "test_curator";

  private static final String NO_ACCESS_USER_PASSWORD = "no_access_pass";
  private static final String NO_ACCESS_USER_LOGIN = "no_access_user";

  @Autowired
  private IDataOverlayService dataOverlayService;

  @Autowired
  private IUserService userService;

  private Project project;
  private ModelData map;
  private Reaction reaction;
  private Element element;
  private User curator;
  private User user;
  private User noAccessUser;

  @Autowired
  private ProjectSnippets snippets;

  @Before
  public void setup() throws Exception {
    project = createAndPersistProject(TEST_PROJECT);
    map = project.getTopModel().getModelData();
    reaction = map.getReactions().iterator().next();
    element = map.getElements().iterator().next();
    curator = createCurator(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD, project);
    user = createUser(TEST_USER_LOGIN, TEST_USER_PASSWORD, project);
    noAccessUser = createUser(NO_ACCESS_USER_LOGIN, NO_ACCESS_USER_PASSWORD);
    assertNotNull(project);
  }

  @After
  public void tearDown() throws Exception {
    removeProject(project);
    removeUser(curator);
    removeUser(user);
    removeUser(noAccessUser);
  }

  @Test
  public void testListOverlaysByCreator() throws Exception {

    createOverlay(TEST_PROJECT, user);

    createOverlay(TEST_PROJECT, curator);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/?creator=" + TEST_CURATOR_LOGIN, TEST_PROJECT)
        .session(session);

    String response = mockMvc.perform(request)
        .andDo(document("projects/project_overlays/get_all_by_creator",
            getProjectPathParameters(),
            getOverlayFilter(),
            listOfOverlayFields()))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    assertEquals("Curator has only one overlay", 1, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  @Test
  public void testListOverlaysByPublicFlag() throws Exception {
    DataOverlay layout = createOverlay(TEST_PROJECT, curator);
    layout.setPublic(true);

    DataOverlay curatorLayout = createOverlay(TEST_PROJECT, curator);
    curatorLayout.setPublic(false);
    dataOverlayService.updateDataOverlay(layout);
    dataOverlayService.updateDataOverlay(curatorLayout);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/?publicOverlay=true")
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    assertEquals("There is only one public overlay", 1, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  @Test
  public void testCuratorShouldSeeAllOverlays() throws Exception {
    DataOverlay layout = createOverlay(TEST_PROJECT, project.getOwner());
    layout.setPublic(true);

    DataOverlay userLayout = createOverlay(TEST_PROJECT, user);
    userLayout.setPublic(false);
    dataOverlayService.updateDataOverlay(layout);
    dataOverlayService.updateDataOverlay(userLayout);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/", TEST_PROJECT)
        .session(session);

    String response = mockMvc.perform(request)
        .andDo(document("projects/project_overlays/get_all",
            getProjectPathParameters(),
            getOverlayFilter(),
            listOfOverlayFields()))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    assertEquals("Curator should see all overlays", 2, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  @Test
  public void testUserShouldSeeAllPublicOverlaysAndOverlays() throws Exception {
    DataOverlay layout = createOverlay(TEST_PROJECT,
        userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    layout.setPublic(true);

    DataOverlay hiddenLayout = createOverlay(TEST_PROJECT,
        userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    hiddenLayout.setPublic(false);

    DataOverlay userLayout = createOverlay(TEST_PROJECT, user);
    userLayout.setPublic(false);

    dataOverlayService.updateDataOverlay(layout);
    dataOverlayService.updateDataOverlay(hiddenLayout);
    dataOverlayService.updateDataOverlay(userLayout);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/")
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    assertEquals("User should see his own overlays and public ones", 2, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  @Test
  public void testNoPrivilegesShouldResultInForbidden() throws Exception {
    MockHttpSession session = createSession(NO_ACCESS_USER_LOGIN, NO_ACCESS_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testAdminCanAccessUserOverlay() throws Exception {
    DataOverlay overlay = createOverlay(TEST_PROJECT, user);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}/", TEST_PROJECT, overlay.getId())
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/get_by_id",
            getOverlayPathParameters(),
            responseFields(overlayFields())))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testCuratorCanAccessUserOverlay() throws Exception {
    DataOverlay overlay = createOverlay(TEST_PROJECT, user);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + "/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCanAccessHisOwnOverlay() throws Exception {
    DataOverlay overlay = createOverlay(TEST_PROJECT, user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + "/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCannotAccessAnotherUserOverlay() throws Exception {
    DataOverlay overlay = createOverlay(TEST_PROJECT, curator);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + "/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testGetBioEntitiesForOverlay() throws Exception {
    ModelData submap = map.getSubmodels().iterator().next().getSubmodel();
    String content = createOverlayContentForAllEntities(map);
    DataOverlay overlay = createOverlay(project, user, content);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), "*")
            .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_overlays/get_bio_entities",
            getOverlayMapPathParameters(),
            listOfOverlayBioEntitiesFields()))
        .andReturn().getResponse().getContentAsString();

    int count = map.getElements().size() + map.getReactions().size()
        + submap.getElements().size() + submap.getReactions().size();
    assertEquals("All bioEntities should be returned", count, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  @Test
  public void testGetBioEntitiesForSpecificSubmap() throws Exception {
    List<ModelData> models = new ArrayList<>(project.getModels());
    ModelData map1 = models.get(0);
    ModelData map2 = models.get(1);
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder requestAll = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), "*")
            .session(session);

    RequestBuilder request1 = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), map1.getId())
            .session(session);

    RequestBuilder request2 = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), map2.getId())
            .session(session);

    String responseAll = mockMvc.perform(requestAll)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    String response1 = mockMvc.perform(request1)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    String response2 = mockMvc.perform(request2)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int countAll = new JsonParser().parse(responseAll).getAsJsonArray().size();
    int count1 = new JsonParser().parse(response1).getAsJsonArray().size();
    int count2 = new JsonParser().parse(response2).getAsJsonArray().size();

    assertTrue(String.format("Too many results %d < %d + %d", countAll, count1, count2), countAll >= count1 + count2);
  }

  @Test
  public void testGetBioEntitiesForOverlayAsUser() throws Exception {
    ModelData submap = map.getSubmodels().iterator().next().getSubmodel();
    String content = createOverlayContentForAllEntities(map);
    DataOverlay overlay = createOverlay(project, user, content);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT,
        overlay.getId(),
        "*")
            .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int count = map.getElements().size() + map.getReactions().size()
        + submap.getElements().size() + submap.getReactions().size();
    assertEquals("All bioEntities should be returned", count, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  private String createOverlayContentForAllEntities(final ModelData map) {
    String content = "element_identifier\tvalue\n";
    for (ModelSubmodelConnection submapConnection : map.getSubmodels()) {
      ModelData submap = submapConnection.getSubmodel();
      for (Reaction entity : submap.getReactions()) {
        content += entity.getElementId() + "\t" + "-1\n";
      }
      for (Element entity : submap.getElements()) {
        content += entity.getElementId() + "\t" + "1\n";
      }
    }
    for (Reaction entity : map.getReactions()) {
      content += entity.getElementId() + "\t" + "-1\n";
    }
    for (Element entity : map.getElements()) {
      content += entity.getElementId() + "\t" + "1\n";
    }
    return content;
  }

  @Test
  public void testGetBioEntitiesForInvalidOverlayId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/ /models/ /bioEntities/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testGetBioEntitiesAsAnonymousForInvalidOverlayId() throws Exception {
    RequestBuilder request = get("/api/projects/" + BUILT_IN_PROJECT + "/overlays/ /models/ /bioEntities/");

    // it should be either 404 or 403 - depending how permissions are checked on
    // invalid overlayId
    mockMvc.perform(request)
        .andExpect(status().is4xxClientError());
  }

  @Test
  public void testGetBioEntitiesForOverlayAsCurator() throws Exception {
    ModelData submap = map.getSubmodels().iterator().next().getSubmodel();
    String content = createOverlayContentForAllEntities(map);
    DataOverlay overlay = createOverlay(project, user, content);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT,
        overlay.getId(),
        "*")
            .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int count = map.getElements().size() + map.getReactions().size()
        + submap.getElements().size() + submap.getReactions().size();
    assertEquals("All bioEntities should be returned", count, new JsonParser()
        .parse(response)
        .getAsJsonArray().size());
  }

  @Test
  public void testGetReactionForOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/{reactionId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), reaction.getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/get_reaction",
            getOverlayReactionPathParameters(),
            listOfOverlayBioEntitiesFields()))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();
  }

  @Test
  public void testGetReactionForOverlayWithWidth() throws Exception {
    String reactionId = reaction.getElementId();
    DataOverlay overlay = createOverlay(project, user, "element_identifier\tvalue\tline_width\n" + reactionId + "\t1\t6");

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/{reactionId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), reaction.getId())
            .session(session);

    String json = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();
    List<Map<String, Object>> objects = objectMapper.readValue(json, new TypeReference<List<Map<String, Object>>>() {
    });
    Map<?, ?> content = (Map<?, ?>) objects.get(0).get("overlayContent");
    assertTrue(content.containsKey("width"));

  }

  private PathParametersSnippet getOverlayReactionPathParameters() {
    return getOverlayMapPathParameters().and(parameterWithName("reactionId").description("reaction identifier"));
  }

  @Test
  public void testGetReactionForOverlayWithColor() throws Exception {
    DataOverlay overlay = createOverlayWithColor(project, user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/{reactionId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), reaction.getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("{method-name}",
            getOverlayReactionPathParameters(),
            listOfOverlayBioEntitiesFields()))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();
  }

  @Test
  public void testGetReactionsForOverlayAsUser() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/{reactionId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), reaction.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetReactionsForOverlayAsCurator() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/{reactionId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), reaction.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetElementsForOverlay() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/elements/{elementId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), element.getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/get_element",
            getOverlayMapPathParameters().and(parameterWithName("elementId").description("element identifier")),
            listOfOverlayBioEntitiesFields()))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetElementsForOverlayAsUser() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/elements/{elementId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), element.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetElementsForOverlayAsCurator() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/elements/{elementId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), element.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testAdminCreateOverlay() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("fileId", String.valueOf(file.getId())),
        new BasicNameValuePair("name", "overlay name"),
        new BasicNameValuePair("description", "overlay name"),
        new BasicNameValuePair("filename", "overlay name"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));

    RequestBuilder request = post("/api/projects/{projectId}/overlays/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/create_from_file",
            getProjectPathParameters(),
            getCreateOverlayRequestParameters(),
            responseFields(overlayFields())))
        .andExpect(status().is2xxSuccessful());

    assertEquals(1, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void testAdminCreateOverlayFromContent() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("content", "element_identifier\tvalue\n\t-1"),
        new BasicNameValuePair("name", "overlay name"),
        new BasicNameValuePair("description", "overlay description"),
        new BasicNameValuePair("filename", "source.txt"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));
    logger.debug(body);

    RequestBuilder request = post("/api/projects/{projectId}/overlays/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/create_from_content",
            getProjectPathParameters(),
            getCreateOverlayRequestParameters(),
            responseFields(overlayFields())))
        .andExpect(status().is2xxSuccessful());

    assertEquals(1, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void testCuratorCreateOverlay() throws Exception {
    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", curator);

    try {
      MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("fileId", String.valueOf(file.getId())),
          new BasicNameValuePair("name", "overlay name"),
          new BasicNameValuePair("description", "overlay name"),
          new BasicNameValuePair("filename", "overlay name"),
          new BasicNameValuePair("googleLicenseConsent", "true"),
          new BasicNameValuePair("type", "GENERIC"))));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT + "/overlays/")
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(session);

      mockMvc.perform(request)
          .andExpect(status().is2xxSuccessful());

      assertEquals(1, dataOverlayService.getDataOverlaysByProject(project).size());
    } finally {
      removeFile(file);
    }
  }

  @Test
  public void testUserCreateOverlay() throws Exception {
    userService.grantUserPrivilege(user, PrivilegeType.CAN_CREATE_OVERLAYS);

    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("fileId", String.valueOf(file.getId())),
        new BasicNameValuePair("name", "overlay name"),
        new BasicNameValuePair("description", "overlay name"),
        new BasicNameValuePair("filename", "overlay name"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT + "/overlays/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    assertEquals(1, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void testUserWithoutCreateOverlayAccessCannotCreateOverlay() throws Exception {
    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", user);

    try {
      MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("fileId", String.valueOf(file.getId())),
          new BasicNameValuePair("name", "overlay name"),
          new BasicNameValuePair("description", "overlay name"),
          new BasicNameValuePair("filename", "overlay name"),
          new BasicNameValuePair("googleLicenseConsent", "false"),
          new BasicNameValuePair("type", "GENERIC"))));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT + "/overlays/")
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(session);

      mockMvc.perform(request)
          .andExpect(status().isForbidden());
    } finally {
      removeFile(file);
    }
  }

  @Test
  public void testUserWithoutAccessToProjectCannotCreateOverlay() throws Exception {
    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("fileId", String.valueOf(file.getId())),
        new BasicNameValuePair("name", "overlay name"),
        new BasicNameValuePair("description", "overlay name"),
        new BasicNameValuePair("filename", "overlay name"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT + "/overlays/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testAdminCanRemoveOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/api/projects/{projectId}/overlays/{overlayId}", TEST_PROJECT, overlay.getId())
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/delete_overlay",
            getOverlayPathParameters()))
        .andExpect(status().is2xxSuccessful());
    assertEquals(0, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void testCuratorCanRemoveOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
    assertEquals(0, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void testUserCanRemoveOwnOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
    assertEquals(0, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void testUserCannotRemoveOtherUserOverlays() throws Exception {
    DataOverlay overlay = createOverlay(curator);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testAdminCanUpdateOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);

    Map<String, Object> updateOverlay = objectMapper.readValue(objectMapper.writeValueAsString(overlay), new TypeReference<Map<String, Object>>() {
    });
    updateOverlay.remove("idObject");
    updateOverlay.remove("defaultOverlay");
    updateOverlay.remove("images");
    updateOverlay.remove("genomeType");
    updateOverlay.remove("genomeVersion");
    updateOverlay.remove("inputDataAvailable");

    String body = "{\"overlay\":" + objectMapper.writeValueAsString(updateOverlay) + "}";

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = patch("/api/projects/{projectId}/overlays/{overlayId}", TEST_PROJECT, overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/update_overlay",
            getOverlayPathParameters(),
            getUpdateOverlayRequestParameters(),
            responseFields(overlayFields())))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testCuratorCanUpdateOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);
    String body = "{\"overlay\":{}}";

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCanUpdateOwnOverlay() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);
    String body = "{\"overlay\":{}}";

    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUpdateInvalidName() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);
    String body = "{\"overlay\":{\"name\":\"\"}}";

    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testUserCannotUpdateOtherUserOverlays() throws Exception {
    DataOverlay overlay = createOverlay(curator);
    String body = "{\"overlay\":{}}";

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testAdminCanAccessOverlaySource() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}:downloadSource", TEST_PROJECT,
        overlay.getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/download_source",
            getOverlayPathParameters()))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testCuratorCanAccessOverlaySource() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_CURATOR_LOGIN, TEST_CURATOR_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + ":downloadSource")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCanAccessOwnOverlaySource() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + ":downloadSource")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCannotAccessOtherUserOverlaysSource() throws Exception {
    DataOverlay overlay = createOverlay(curator);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + ":downloadSource")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testOverlayCreatedAsFirstShouldHaveProperOrderIndex() throws Exception {
    createOverlay(user);

    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", admin);
    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("fileId", String.valueOf(file.getId())),
        new BasicNameValuePair("name", "overlay name"),
        new BasicNameValuePair("description", "overlay name"),
        new BasicNameValuePair("filename", "overlay name"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT + "/overlays/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int orderNumber = new JsonParser()
        .parse(response)
        .getAsJsonObject()
        .get("order")
        .getAsInt();

    assertEquals("First user data overlay should be ordered with 1", 1, orderNumber);
  }

  @Test
  public void testUserCanAccessPublicOverlay() throws Exception {
    DataOverlay overlay = createOverlay(TEST_PROJECT, project.getOwner());
    overlay.setPublic(true);
    dataOverlayService.updateDataOverlay(overlay);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId() + "/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCanAccessDataInPublicOverlay() throws Exception {
    DataOverlay overlay = createOverlay(project.getOwner());
    overlay.setPublic(true);
    dataOverlayService.updateDataOverlay(overlay);
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}/models/*/bioEntities/",
        TEST_PROJECT,
        overlay.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUserCanAccessReactionDataInPublicOverlay() throws Exception {
    DataOverlay overlay = createOverlay(project.getOwner());
    overlay.setPublic(true);
    dataOverlayService.updateDataOverlay(overlay);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/-1/",
        TEST_PROJECT,
        overlay.getId(),
        map.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testUserCanAccessElementDataInPublicOverlay() throws Exception {
    DataOverlay overlay = createOverlay(project.getOwner());
    overlay.setPublic(true);
    dataOverlayService.updateDataOverlay(overlay);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/elements/-1/",
        TEST_PROJECT,
        overlay.getId(),
        map.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testUserCanDownloadSourceOfPublicOverlay() throws Exception {
    DataOverlay overlay = createOverlay(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    overlay.setPublic(true);
    dataOverlayService.updateDataOverlay(overlay);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}:downloadSource/",
        TEST_PROJECT,
        overlay.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  private DataOverlay createOverlay(final User user) throws Exception {
    return createOverlay(TEST_PROJECT, user);
  }

  @Test
  public void testListOverlaysWithUnknownProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/overlays/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetOverlayByIdWithUnknownProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/overlays/" + overlay.getId() + "/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetBioEntitiesForOverlayWithUndefinedProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/overlays/{overlayId}/models/{mapId}/bioEntities/",
        overlay.getId(),
        map.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetReactionsForOverlayWithUndefinedProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/*/overlays/{overlayId}/models/{mapId}/bioEntities/reactions/{reactionId}/",
        overlay.getId(),
        map.getId(),
        reaction.getId())
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetElementsForOverlayWithUndefinedProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/overlays/" + overlay.getId() + "/models/" + map.getId()
        + "/bioEntities/elements/" + element.getId() + "/")
            .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testCreateOverlayWithUndefinedProjectId() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("fileId", String.valueOf(file.getId())),
        new BasicNameValuePair("name", "overlay name"),
        new BasicNameValuePair("description", "overlay name"),
        new BasicNameValuePair("filename", "overlay name"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));

    RequestBuilder request = post("/api/projects/*/overlays/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testRemoveOverlayWithUndefinedProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/api/projects/*/overlays/" + overlay.getId())
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testUpdateOverlayWithUndefinedProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    String body = "{\"overlay\":{}}";

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = patch("/api/projects/*/overlays/" + overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testAccessOverlaySourceWithUndefinedProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay = createOverlay(admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/overlays/" + overlay.getId() + ":downloadSource")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testOrderAfterRemoveOverlay() throws Exception {
    DataOverlay overlay1 = createOverlay(user);
    overlay1.setOrderIndex(1);
    DataOverlay overlay2 = createOverlay(user);
    overlay2.setOrderIndex(2);
    DataOverlay overlay3 = createOverlay(user);
    overlay3.setOrderIndex(3);

    dataOverlayService.updateDataOverlay(overlay1);
    dataOverlayService.updateDataOverlay(overlay2);
    dataOverlayService.updateDataOverlay(overlay3);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay2.getId())
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
    assertEquals(1, dataOverlayService.getDataOverlayById(overlay1.getId()).getOrderIndex());
    assertEquals(2, dataOverlayService.getDataOverlayById(overlay3.getId()).getOrderIndex());
  }

  @Test
  public void testOrderAfterRemoveOverlayForPublicOverlay() throws Exception {
    DataOverlay overlay1 = createOverlay(project.getOwner());
    overlay1.setPublic(true);
    overlay1.setOrderIndex(1);
    DataOverlay overlay2 = createOverlay(project.getOwner());
    overlay2.setPublic(true);
    overlay2.setOrderIndex(2);
    DataOverlay overlay3 = createOverlay(project.getOwner());
    overlay3.setPublic(true);
    overlay3.setOrderIndex(3);

    dataOverlayService.updateDataOverlay(overlay1);
    dataOverlayService.updateDataOverlay(overlay2);
    dataOverlayService.updateDataOverlay(overlay3);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay2.getId())
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
    assertEquals(1, dataOverlayService.getDataOverlayById(overlay1.getId()).getOrderIndex());
    assertEquals(2, dataOverlayService.getDataOverlayById(overlay3.getId()).getOrderIndex());
  }

  @Test
  public void testChangeOrderAfterChaningOwner() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    DataOverlay overlay1 = createOverlay(user);
    overlay1.setOrderIndex(1);
    DataOverlay overlay2 = createOverlay(user);
    overlay2.setOrderIndex(2);

    DataOverlay overlay3 = createOverlay(admin);
    overlay3.setOrderIndex(1);
    DataOverlay overlay4 = createOverlay(admin);
    overlay4.setOrderIndex(2);

    dataOverlayService.updateDataOverlay(overlay1);
    dataOverlayService.updateDataOverlay(overlay2);
    dataOverlayService.updateDataOverlay(overlay3);
    dataOverlayService.updateDataOverlay(overlay4);

    String body = "{\"overlay\":{\"creator\":\"" + admin.getLogin() + "\"}}";

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay1.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    assertEquals("Order of overlays wasn't updated for original owner of overlay", 1,
        dataOverlayService.getDataOverlayById(overlay2.getId()).getOrderIndex());

    assertTrue(dataOverlayService.getDataOverlayById(overlay1.getId()).getOrderIndex() == 1
        || dataOverlayService.getDataOverlayById(overlay3.getId()).getOrderIndex() == 1);
    assertTrue(dataOverlayService.getDataOverlayById(overlay1.getId()).getOrderIndex() == 2
        || dataOverlayService.getDataOverlayById(overlay3.getId()).getOrderIndex() == 2);
    assertEquals(3, dataOverlayService.getDataOverlayById(overlay4.getId()).getOrderIndex());
  }

  @Test
  public void testCreateOverlayWithoutCreateDataOverlayAccess() throws Exception {
    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", user);

    try {
      MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("fileId", String.valueOf(file.getId())),
          new BasicNameValuePair("name", "overlay name"),
          new BasicNameValuePair("description", "overlay name"),
          new BasicNameValuePair("filename", "overlay name"),
          new BasicNameValuePair("googleLicenseConsent", "true"),
          new BasicNameValuePair("type", "GENERIC"))));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT + "/overlays/")
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(session);

      mockMvc.perform(request)
          .andExpect(status().isForbidden());
    } finally {
      removeFile(file);
    }
  }

  protected RequestParametersSnippet getOverlayFilter() {
    return requestParameters(
        parameterWithName("creator")
            .description("login of the user who created the overlay")
            .optional(),
        parameterWithName("publicOverlay")
            .description(
                "true/false - are we interested in public overlays or private custom data sets (default: false)")
            .optional());
  }

  private ResponseFieldsSnippet listOfOverlayFields() {
    return responseFields().andWithPrefix("[].", overlayFields());
  }

  private List<FieldDescriptor> overlayFields() {
    return Arrays.asList(
        fieldWithPath("name")
            .description("name")
            .type(JsonFieldType.STRING),
        fieldWithPath("description")
            .description("description")
            .type(JsonFieldType.STRING),
        fieldWithPath("idObject")
            .description("identifier")
            .type(JsonFieldType.NUMBER),
        fieldWithPath("publicOverlay")
            .description("is the data overlay publicly available to all users")
            .type(JsonFieldType.BOOLEAN),
        fieldWithPath("googleLicenseConsent")
            .description("did creator accepted license required by Google Maps API")
            .type(JsonFieldType.BOOLEAN),
        fieldWithPath("order")
            .description("sort order")
            .type(JsonFieldType.NUMBER),
        fieldWithPath("type")
            .description("type; available options: " + new ProjectSnippets().getOptionsAsString(DataOverlayType.class))
            .type(JsonFieldType.STRING),
        fieldWithPath("genomeType")
            .description("reference genome type; available options: "
                + new ProjectSnippets().getOptionsAsString(ReferenceGenomeType.class))
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath("genomeVersion")
            .description("reference genome version")
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath("creator")
            .description("login of the user who uploaded data overlay")
            .type(JsonFieldType.STRING)
            .optional());
  }

  private Snippet getCreateOverlayRequestParameters() {
    return requestParameters(
        parameterWithName("name")
            .description("name of the data overlay"),
        parameterWithName("content")
            .description(
                "content of the file that is uploaded with definition what should be visible where (alternative to fileId)")
            .optional(),
        parameterWithName("fileId")
            .description(
                "uploaded file id that should be used as data overlay content content (alternative to content)")
            .optional(),
        parameterWithName("description")
            .description("short description of the data overlay"),
        parameterWithName("filename")
            .description("name of the file that should be used when downloading the source"),
        parameterWithName("type")
            .description("type of the data overlay (default: " + DataOverlayType.GENERIC + "). Available options: "
                + snippets.getOptionsAsString(DataOverlayType.class))
            .optional(),
        parameterWithName("googleLicenseConsent")
            .description(
                "true/false indicating if user agreed to the new Google Maps API license https://cloud.google.com/maps-platform/terms/"));
  }

  private Snippet getUpdateOverlayRequestParameters() {
    List<FieldDescriptor> fields = new ArrayList<>(overlayFields());
    List<FieldDescriptor> toRemove = new ArrayList<>();
    for (FieldDescriptor fieldDescriptor : fields) {
      if (fieldDescriptor.getPath().equals("idObject")
          || fieldDescriptor.getPath().equals("inputDataAvailable")
          || fieldDescriptor.getPath().equals("images")
          || fieldDescriptor.getPath().equals("genomeType")
          || fieldDescriptor.getPath().equals("genomeVersion")
          || fieldDescriptor.getPath().equals("defaultOverlay")) {
        toRemove.add(fieldDescriptor);
      }
    }
    fields.removeAll(toRemove);
    return requestFields().andWithPrefix("overlay.", fields);
  }

  private ResponseFieldsSnippet listOfOverlayBioEntitiesFields() {
    return responseFields().andWithPrefix("[].", overlayBioEntityFields());
  }

  private List<FieldDescriptor> overlayBioEntityFields() {
    List<FieldDescriptor> result = new ArrayList<>(Arrays.asList(
        fieldWithPath("type")
            .description("type of bioEntity (ALIAS/REACTION)")
            .type(JsonFieldType.STRING),
        fieldWithPath("overlayContent.idObject")
            .description("identifier of the bioEntity")
            .type(JsonFieldType.STRING),
        fieldWithPath("overlayContent.uniqueId")
            .description("identifier of the bioEntity")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath("overlayContent.modelId")
            .description("map identifier where bioEntity is located")
            .type(JsonFieldType.NUMBER),
        fieldWithPath("overlayContent.value")
            .description("normalized value assigned to the bioEntity from overlay")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath("overlayContent.color")
            .description("color assigned to the bioEntity from overlay")
            .optional()
            .type(JsonFieldType.OBJECT),
        fieldWithPath("overlayContent.description")
            .description("description assigned to the bioEntity from overlay")
            .type(JsonFieldType.STRING)
            .optional(),
        fieldWithPath("overlayContent.geneVariations")
            .description("list of gene variants assigned to the bioEntity from overlay")
            .type(JsonFieldType.ARRAY)
            .optional(),
        fieldWithPath("overlayContent.type")
            .description("type of data overlay")
            .type(JsonFieldType.STRING)
            .optional()
            .ignored(),
        fieldWithPath("overlayContent.width")
            .description("line width of reaction bioEntity assigned to the bioEntity from overlay")
            .type(JsonFieldType.NUMBER)
            .optional()));
    result.addAll(colorField("overlayContent.color", true, false));
    return result;
  }

  private Collection<FieldDescriptor> colorField(final String originalPrefix, final boolean ignore, final boolean required) {
    String prefix = originalPrefix;
    if (!prefix.isEmpty() && !prefix.endsWith(".")) {
      prefix = prefix + ".";
    }
    List<FieldDescriptor> result = Arrays.asList(
        fieldWithPath(prefix + "alpha")
            .description("alpha value (0-255)")
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "rgb")
            .description("RED-GREEN-BLUE value as integer #RRGGBB")
            .type(JsonFieldType.NUMBER));
    if (ignore) {
      for (final FieldDescriptor fieldDescriptor : result) {
        fieldDescriptor.ignored();
      }
    }
    if (!required) {
      for (final FieldDescriptor fieldDescriptor : result) {
        fieldDescriptor.optional();
      }
    }
    return result;
  }

  @Test
  public void testGetBioEntitiesForNonExistingOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/-1/models/*/bioEntities/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetReactionForNonExistingOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/-1/models/-1/bioEntities/reactions/123/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetReactionForInvalidModelId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/-1/models/abc/bioEntities/reactions/123/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testGetElementForInvalidModelId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/-1/models/abc/bioEntities/elements/123/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testGetElementForInvalidOverlayId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/abc/models/-1/bioEntities/elements/123/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testGetElementForNonExistingOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/-1/models/-1/bioEntities/elements/123/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testRemoveNonExistingOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/-1/").session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testUpdateNonExistingOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    String body = "{\"overlay\":{\"overlayId\":-1}}";
    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/-1/")
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testUpdateIvalidOverlayId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    String body = "{\"overlay\":{\"overlayId\":-1}}";
    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/overlays/bla/")
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testFetchSourceForExistingOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/" + TEST_PROJECT + "/overlays/-1:downloadSource").session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testRemoveUserOverlayAsAnonymous() throws Exception {
    DataOverlay overlay = createOverlay(TEST_PROJECT, user);
    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT + "/overlays/" + overlay.getId());

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void testFetchGeneticDataOverlayData() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    DataOverlay overlay = createOverlay(project, user, "#TYPE=GENETIC_VARIANT\n"
        + "#GENOME_TYPE=UCSC\n"
        + "#GENOME_VERSION=xenLae2\n"
        + "position\toriginal_dna\talternative_dna\tgene_name\tdescription\tcolor\tcontig\n"
        + "10146\tAC\tA\tGSTA4\tupstream\t#ff0000\tchr1", DataOverlayType.GENETIC_VARIANT);

    assertNotNull(overlay);
    for (final Element elementIterator : map.getElements()) {
      if (elementIterator.getName().equals("GSTA4")) {
        element = elementIterator;
      }
    }
    assertEquals(element.getName(), "GSTA4");

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/elements/{elementId}/",
        TEST_PROJECT, overlay.getId(), map.getId(), element.getId())
            .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();
    objectMapper.readValue(response, Object.class);
  }

  @Test
  public void testFetchGeneticDataOverlay() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    DataOverlay overlay = createOverlay(project, user, "#TYPE=GENETIC_VARIANT\n"
        + "#GENOME_TYPE=UCSC\n"
        + "#GENOME_VERSION=xenLae2\n"
        + "position\toriginal_dna\talternative_dna\tgene_name\tdescription\tcolor\tcontig\n"
        + "10146\tAC\tA\tGSTA4\tupstream\t#ff0000\tchr1", DataOverlayType.GENETIC_VARIANT);

    assertNotNull(overlay);
    for (final Element elementIterator : map.getElements()) {
      if (elementIterator.getName().equals("GSTA4")) {
        element = elementIterator;
      }
    }
    assertEquals(element.getName(), "GSTA4");

    RequestBuilder request = get(
        "/api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), map.getId())
            .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();
    objectMapper.readValue(response, Object.class);
  }

  @Test
  public void testDownloadOverlayLegend() throws Exception {
    DataOverlay overlay = createOverlay(user);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/overlays/{overlayId}:downloadLegend", TEST_PROJECT,
        overlay.getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_overlays/download_legend",
            getOverlayPathParameters()))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testUpdateOverlayNameTooLong() throws Exception {
    DataOverlay overlay = createOverlay(user);

    Map<String, Object> updateOverlay = new HashMap<>();
    updateOverlay.put("name", INVALID_LONG_NAME);

    String body = "{\"overlay\":" + new Gson().toJson(updateOverlay) + "}";

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = patch("/api/projects/{projectId}/overlays/{overlayId}", TEST_PROJECT, overlay.getId())
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testInvalidName() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    UploadedFileEntry file = createFile("element_identifier\tvalue\n\t-1", admin);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("fileId", String.valueOf(file.getId())),
        new BasicNameValuePair("name", " "),
        new BasicNameValuePair("description", "overlay name"),
        new BasicNameValuePair("filename", "overlay name"),
        new BasicNameValuePair("googleLicenseConsent", "false"),
        new BasicNameValuePair("type", "GENERIC"))));

    RequestBuilder request = post("/api/projects/{projectId}/overlays/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testUpdateWithEmptyBody() throws Exception {
    DataOverlay overlay = createOverlay(user);

    String body = "{}";

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = patch("/api/projects/{projectId}/overlays/{overlayId}", TEST_PROJECT, overlay.getId())
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

}
