package lcsb.mapviewer.web;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.delete;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.patch;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.post;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.requestFields;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.payload.PayloadDocumentation.subsectionWithPath;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.requestParameters;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.JsonFieldType;
import org.springframework.restdocs.request.PathParametersSnippet;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.gson.JsonParser;

import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.ProjectStatus;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.graphics.MapCanvasType;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.layout.ProjectBackground;
import lcsb.mapviewer.model.map.layout.ProjectBackgroundStatus;
import lcsb.mapviewer.model.security.Privilege;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.ConfigurationElementType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.interfaces.IConfigurationService;
import lcsb.mapviewer.services.interfaces.IProjectBackgroundService;
import lcsb.mapviewer.services.interfaces.IProjectService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.web.serialization.ProjectBackgroundUpdateMixIn;

@RunWith(SpringJUnit4ClassRunner.class)
public class ProjectControllerIntegrationTest extends ControllerIntegrationTest {

  private static final String CURATOR_PASSWORD = "curator_pass";
  private static final String CURATOR_LOGIN = "curator_user";

  private static final String USER_PASSWORD = "user_pass";
  private static final String USER_LOGIN = "test_user";

  private static final String TEST_PROJECT_2 = "test_project2";

  @Autowired
  private IUserService userService;

  @Autowired
  private IProjectService projectService;

  @Autowired
  private IProjectBackgroundService projectBackgroundService;

  @Autowired
  private ProjectSnippets snippets;

  @Autowired
  private ObjectMapper objectMapper;

  private User curator;

  @Autowired
  private IConfigurationService configurationService;

  @Before
  public void setup() {
    curator = createCurator(CURATOR_LOGIN, CURATOR_PASSWORD);
    configurationService.setConfigurationValue(ConfigurationElementType.MINERVA_ROOT, "https://pdmap.uni.lu/minerva/");
  }

  @After
  public void tearDown() throws Exception {
    removeProject(TEST_PROJECT);
    removeProject(TEST_PROJECT_2);
    removeUser(curator);
    removeUser(userService.getUserByLogin(USER_LOGIN));
    configurationService.setConfigurationValue(ConfigurationElementType.MINERVA_ROOT, "");
  }

  @Test
  public void testGetAllProjectsAsAdmin() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    RequestBuilder request = get("/api/projects/")
        .session(session);

    Project project = projectService.getProjectByProjectId(BUILT_IN_PROJECT);
    project.setDirectory(project.getDirectory().replaceAll("[a-z]", "0"));
    projectService.updateProject(project);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_data/list",
            responseFields(
                subsectionWithPath("[]")
                    .description("list of projects").type(JsonFieldType.ARRAY))))
        .andReturn().getResponse().getContentAsString();

    int projects = new JsonParser()
        .parse(response)
        .getAsJsonArray().size();

    assertEquals("admin should see all projects", projectService.getAllProjects().size(), projects);
  }

  @Test
  public void testGetSourceFile() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    createAndPersistProject(TEST_PROJECT);
    RequestBuilder request = get("/api/projects/{projectId}:downloadSource", TEST_PROJECT)
        .session(session);
    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_data/get_source_file",
            projectPathParameters()));

  }

  @Test
  public void testGetLimitedProjectsAsCurator() throws Exception {
    Project project = createAndPersistProject(TEST_PROJECT);

    userService.grantUserPrivilege(curator, PrivilegeType.READ_PROJECT, project.getProjectId());
    Project project2 = new Project(TEST_PROJECT_2);
    project2.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    projectService.add(project2);

    MockHttpSession session = createSession(CURATOR_LOGIN, CURATOR_PASSWORD);

    RequestBuilder request = get("/api/projects/")
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int projects = new JsonParser()
        .parse(response)
        .getAsJsonArray().size();

    assertEquals("curator should see only limited list of projects", 1, projects);
  }

  @Test
  public void testGetLimitedProjectsAsGuest() throws Exception {
    Project project2 = new Project(TEST_PROJECT_2);
    project2.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    projectService.add(project2);

    RequestBuilder request = get("/api/projects/");

    String response = mockMvc.perform(request)
        .andDo(document("projects/project_data/list_guest",
            responseFields(
                subsectionWithPath("[]")
                    .description("list of projects").type(JsonFieldType.ARRAY))))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int projects = new JsonParser()
        .parse(response)
        .getAsJsonArray().size();

    assertEquals("guest should see only limited list of projects", 1, projects);
  }

  @Test
  public void testUserPrivilegesChangeDuringActiveSession() throws Exception {
    MockHttpSession session = createSession(CURATOR_LOGIN, CURATOR_PASSWORD);

    Project project = createAndPersistProject(TEST_PROJECT);

    userService.grantUserPrivilege(curator, PrivilegeType.READ_PROJECT, project.getProjectId());

    RequestBuilder request = get("/api/projects/{projectId}/", TEST_PROJECT)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_data/get_project",
            projectPathParameters(),
            snippets.getProjectSnippet()))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetLogsForProject() throws Exception {
    Project project = createAndPersistProject(TEST_PROJECT);

    userService.grantUserPrivilege(curator, PrivilegeType.READ_PROJECT, project.getProjectId());

    MockHttpSession session = createSession(CURATOR_LOGIN, CURATOR_PASSWORD);

    RequestBuilder request = get(
        "/api/projects/{projectId}/logs/?start=0&length=5&level=warning&sortColumn=id&sortOrder=asc&search=age",
        TEST_PROJECT).session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_data/get_project_logs",
            requestParameters(
                parameterWithName("start").description("number of first entry in this response").optional(),
                parameterWithName("sortColumn")
                    .description("log entry column that should be used for sorting (default: id)").optional(),
                parameterWithName("sortOrder").description("log entry sort order (asc, desc)").optional(),
                parameterWithName("level").description("level of log entry (warning, error)").optional(),
                parameterWithName("length").description("number of log entries we want to obtain").optional(),
                parameterWithName("search").description("search query used for filtering").optional()),
            projectPathParameters(),
            responseFields(
                subsectionWithPath("data")
                    .description("list of log entries")
                    .type(JsonFieldType.ARRAY),
                fieldWithPath("filteredSize")
                    .description("number of entries that match filter criteria")
                    .type(JsonFieldType.NUMBER),
                fieldWithPath("length")
                    .description("number of entries in this response")
                    .type(JsonFieldType.NUMBER),
                fieldWithPath("start")
                    .description("number of first entry in this response")
                    .type(JsonFieldType.NUMBER),
                fieldWithPath("totalSize")
                    .description("number of all log entries")
                    .type(JsonFieldType.NUMBER))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetStatisticsForProject() throws Exception {
    createAndPersistProject(TEST_PROJECT);
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/statistics", TEST_PROJECT).session(session);

    mockMvc.perform(request).andDo(document("projects/project_data/get_project_statistics",
        projectPathParameters(),
        responseFields(
            subsectionWithPath("reactionAnnotations")
                .description("list of reaction annotation types")
                .type(JsonFieldType.OBJECT),
            subsectionWithPath("elementAnnotations")
                .description("list of element annotation types")
                .type(JsonFieldType.OBJECT),
            fieldWithPath("publications")
                .description("number of publications included in the project")
                .type(JsonFieldType.NUMBER))))
        .andExpect(status().is2xxSuccessful());
    assertEquals(0, getErrorLogs().size());
  }

  @Test
  public void testGrantPrivilege() throws Exception {
    createAndPersistProject(TEST_PROJECT);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "[{"
        + "\"privilegeType\":\"" + PrivilegeType.READ_PROJECT + "\", "
        + "\"login\":\"" + CURATOR_LOGIN + "\""
        + "}]";

    RequestBuilder request = patch("/api/projects/{projectId}:grantPrivileges", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_privileges/grant_privilege",
            projectPathParameters(),
            requestFields(
                fieldWithPath("[]")
                    .description("list of privileges to grant")
                    .type(JsonFieldType.ARRAY),
                fieldWithPath("[].privilegeType")
                    .description("type of privilege (READ_PROJECT/WRITE_PROJECT)")
                    .type(JsonFieldType.STRING),
                fieldWithPath("[].login")
                    .description("user login who should gain access")
                    .type(JsonFieldType.STRING))))
        .andExpect(status().is2xxSuccessful());

    assertTrue(userService.getUserByLogin(CURATOR_LOGIN, true).getPrivileges()
        .contains(new Privilege(PrivilegeType.READ_PROJECT, TEST_PROJECT)));
  }

  @Test
  public void testRevokePrivilege() throws Exception {
    Project project = createAndPersistProject(TEST_PROJECT);
    userService.grantUserPrivilege(curator, PrivilegeType.READ_PROJECT, project.getProjectId());

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "[{"
        + "\"privilegeType\":\"" + PrivilegeType.READ_PROJECT + "\", "
        + "\"login\":\"" + CURATOR_LOGIN + "\""
        + "}]";

    RequestBuilder request = patch("/api/projects/{projectId}:revokePrivileges", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_privileges/revoke_privilege",
            projectPathParameters(),
            requestFields(
                fieldWithPath("[]")
                    .description("list of privileges to revoke")
                    .type(JsonFieldType.ARRAY),
                fieldWithPath("[].privilegeType")
                    .description("type of privilege (READ_PROJECT/WRITE_PROJECT)")
                    .type(JsonFieldType.STRING),
                fieldWithPath("[].login")
                    .description("user login who should lose access")
                    .type(JsonFieldType.STRING))))
        .andExpect(status().is2xxSuccessful());

    assertFalse(curator.getPrivileges().contains(new Privilege(PrivilegeType.READ_PROJECT, TEST_PROJECT)));
  }

  @Test
  public void testGetNonExistingProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testUpdateProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    createAndPersistProject(TEST_PROJECT);

    String content = "{\"project\":{"
        + "\"version\":\"xxx\", "
        + "\"name\": \"new name\", "
        + "\"notifyEmail\": \"notify@uni.lu\", "
        + "\"organism\": {\"type\": \"TAXONOMY\", \"resource\": \"9605\"}, "
        + "\"disease\": {\"type\": \"MESH_2012\", \"resource\": \"D010300\"}, "
        + "\"mapCanvasType\": \"OPEN_LAYERS\" "
        + "}}";

    RequestBuilder request = patch("/api/projects/{projectId}/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_JSON)
        .content(content)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_data/update",
            projectPathParameters(),
            requestFields(
                fieldWithPath("project.version")
                    .description("version of the project")
                    .type(JsonFieldType.STRING)
                    .optional(),
                fieldWithPath("project.name")
                    .description("name of the project")
                    .type(JsonFieldType.STRING)
                    .optional(),
                fieldWithPath("project.notifyEmail")
                    .description("email address that should be when something change in the project")
                    .type(JsonFieldType.STRING)
                    .optional(),
                subsectionWithPath("project.organism")
                    .description("organism associated with the project")
                    .type(JsonFieldType.OBJECT)
                    .optional(),
                subsectionWithPath("project.disease")
                    .description("disease associated with the project")
                    .type(JsonFieldType.OBJECT)
                    .optional(),
                fieldWithPath("project.mapCanvasType")
                    .description(
                        "type of map canvas engine to be used when visualizing map, for now there are only two options available: "
                            + snippets.getOptionsAsString(MapCanvasType.class))
                    .type(JsonFieldType.STRING)
                    .optional()),
            snippets.getProjectSnippet()))

        .andExpect(status().is2xxSuccessful());

    assertEquals("xxx", projectService.getProjectByProjectId(TEST_PROJECT).getVersion());
    assertEquals("9605", projectService.getProjectByProjectId(TEST_PROJECT).getOrganism().getResource());
    assertEquals(MiriamType.TAXONOMY, projectService.getProjectByProjectId(TEST_PROJECT).getOrganism().getDataType());
  }

  @Test
  public void testUpdateWithEmptyDiseaseProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    createAndPersistProject(TEST_PROJECT);

    String content = "{\"project\":{"
        + "\"version\":\"xxx\", "
        + "\"name\": \"new name\", "
        + "\"notifyEmail\": \"notify@uni.lu\", "
        + "\"organism\": {}, "
        + "\"disease\": {}, "
        + "\"mapCanvasType\": \"OPEN_LAYERS\" "
        + "}}";

    RequestBuilder request = patch("/api/projects/{projectId}/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_JSON)
        .content(content)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    assertEquals(null, projectService.getProjectByProjectId(TEST_PROJECT).getOrganism());
    assertEquals(null, projectService.getProjectByProjectId(TEST_PROJECT).getDisease());

  }

  @Test
  public void testUpdateProjectWithTooLongVersion() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    createAndPersistProject(TEST_PROJECT);

    String content = "{\"project\":{\"version\":\"12345678901234567890123456\"}}";

    RequestBuilder request = patch("/api/projects/" + TEST_PROJECT + "/")
        .contentType(MediaType.APPLICATION_JSON)
        .content(content)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testUpdateProjectWithUndefinedProjectId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String content = "{\"project\":{\"version\":\"xxx\"}}";

    RequestBuilder request = patch("/api/projects/*/")
        .contentType(MediaType.APPLICATION_JSON)
        .content(content)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGrantPrivilegeForUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "[{"
        + "\"privilegeType\":\"" + PrivilegeType.READ_PROJECT + "\", "
        + "\"login\":\"" + CURATOR_LOGIN + "\""
        + "}]";

    RequestBuilder request = patch("/api/projects/*:grantPrivileges")
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testRevokePrivilegeForUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "[{"
        + "\"privilegeType\":\"" + PrivilegeType.READ_PROJECT + "\", "
        + "\"login\":\"" + CURATOR_LOGIN + "\""
        + "}]";

    RequestBuilder request = patch("/api/projects/*:revokePrivileges")
        .contentType(MediaType.APPLICATION_JSON)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testRemoveProjectForUndefinedProjectId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/api/projects/*/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetStatisticsForUndefinedProjectId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/statistics")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testDownloadSourceForUndefinedProjectId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*:downloadSource")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetLogsForUndefinedProjectId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/logs/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetSubmapConnectionsForUndefinedProjectId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/*/submapConnections/").session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetSubmapConnectionsForProject() throws Exception {
    createAndPersistProject(TEST_PROJECT);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/api/projects/{projectId}/submapConnections/", TEST_PROJECT).session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_maps/get_connections",
            projectPathParameters(),
            responseFields(
                fieldWithPath("[]").description("list of connections").type(JsonFieldType.ARRAY),
                fieldWithPath("[].from.id").description("anchor bioEntity id").type(JsonFieldType.NUMBER),
                fieldWithPath("[].from.modelId").description("anchor map id").type(JsonFieldType.NUMBER),
                fieldWithPath("[].from.type").description("anchor type").type(JsonFieldType.STRING),
                fieldWithPath("[].to.modelId").description("destination map id").type(JsonFieldType.NUMBER))));
  }

  @Test
  public void addProjectWithInvalidFileId() throws Exception {
    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", "-1"),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

    mockMvc.perform(request).andExpect(status().isBadRequest());
  }

  @Test
  public void addInvalidProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        admin);
    try {

      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser",
              "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

      RequestBuilder request = post("/api/projects/{projectId}", INVALID_LONG_NAME)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().isBadRequest());

    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void addProjectWithTooLongName() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        admin);
    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("name", INVALID_LONG_NAME),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().isBadRequest());

    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void addProjectWithTooLongVersion() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        admin);
    try {
      String invalidVersion = "12345678901234567890123456";

      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("version", invalidVersion),
          new BasicNameValuePair("parser",
              "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().isBadRequest());

    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void modifyProjectWithTooLongName() throws Exception {
    createAndPersistProject(TEST_PROJECT);

    String content = "{\"project\":{\"name\":\"" + INVALID_LONG_NAME + "\"}}";

    RequestBuilder request = patch("/api/projects/{projectId}/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_JSON)
        .content(content)
        .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

    mockMvc.perform(request).andExpect(status().isBadRequest());
  }

  @Test
  public void removeProjectWithUsedFile() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    try {
      createFullProject(admin);

      RequestBuilder request = delete("/api/projects/{projectId}/", TEST_PROJECT_2)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().is2xxSuccessful())
          .andDo(document("projects/project_data/delete",
              projectPathParameters(),
              snippets.getProjectSnippet()));

    } finally {
      waitForProjectToFinishLoading(TEST_PROJECT);
    }
  }

  private void createFullProject(final User admin)
      throws UnsupportedEncodingException, IOException, Exception, InterruptedException {
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        admin);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("name", "Project name"),
        new BasicNameValuePair("version", "0.0.1"),
        new BasicNameValuePair("notify-email", "minerva@uni.lu"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));
    mockMvc.perform(request).andExpect(status().is2xxSuccessful());

    request = post("/api/projects/" + TEST_PROJECT_2)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));
    mockMvc.perform(request).andExpect(status().is2xxSuccessful());

    waitForProjectToFinishLoading(TEST_PROJECT_2);

    Project project = projectService.getProjectByProjectId(TEST_PROJECT_2);
    project.setDirectory(project.getDirectory().replaceAll("[a-z]", "0"));
    projectService.updateProject(project);
  }

  @Test
  public void addComplexProjectWithInvalidOverlayName() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        Files.readAllBytes(Paths.get("./src/test/resources/complex_model_with_layouts.zip")), admin);
    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"),

          new BasicNameValuePair("zip-entries[0][_type]", "MAP"),
          new BasicNameValuePair("zip-entries[0][_filename]", "main.xml"),
          new BasicNameValuePair("zip-entries[0][_data][root]", "true"),
          new BasicNameValuePair("zip-entries[0][_data][name]", "s1"),
          new BasicNameValuePair("zip-entries[0][_data][type][id]", "UNKNOWN"),
          new BasicNameValuePair("zip-entries[0][_data][type][name]", "Unknown"),

          new BasicNameValuePair("zip-entries[1][_type]", "OVERLAY"),
          new BasicNameValuePair("zip-entries[1][_filename]", "layouts/goodSchema.txt"),
          new BasicNameValuePair("zip-entries[1][_data][name]", "")

      )));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().isBadRequest());
    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void addComplexProject() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        Files.readAllBytes(Paths.get("./src/test/resources/complex_model_with_layouts.zip")), admin);
    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"),

          new BasicNameValuePair("zip-entries[0][_type]", "MAP"),
          new BasicNameValuePair("zip-entries[0][_filename]", "main.xml"),
          new BasicNameValuePair("zip-entries[0][_data][root]", "true"),
          new BasicNameValuePair("zip-entries[0][_data][name]", "s1"),
          new BasicNameValuePair("zip-entries[0][_data][type][id]", "UNKNOWN"),
          new BasicNameValuePair("zip-entries[0][_data][type][name]", "Unknown"),

          new BasicNameValuePair("zip-entries[1][_type]", "OVERLAY"),
          new BasicNameValuePair("zip-entries[1][_filename]", "layouts/goodSchema.txt"),
          new BasicNameValuePair("zip-entries[1][_data][name]", "test-o")

      )));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().is2xxSuccessful());
    } finally {
      waitForProjectToFinishLoading(TEST_PROJECT);
    }
  }

  @Test
  public void addComplexProjectWithInvalidOverlay() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        Files.readAllBytes(Paths.get("./src/test/resources/complex_model_with_invalid_layouts.zip")), admin);
    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"),

          new BasicNameValuePair("zip-entries[0][_type]", "MAP"),
          new BasicNameValuePair("zip-entries[0][_filename]", "main.xml"),
          new BasicNameValuePair("zip-entries[0][_data][root]", "true"),
          new BasicNameValuePair("zip-entries[0][_data][name]", "s1"),
          new BasicNameValuePair("zip-entries[0][_data][type][id]", "UNKNOWN"),
          new BasicNameValuePair("zip-entries[0][_data][type][name]", "Unknown"),

          new BasicNameValuePair("zip-entries[1][_type]", "OVERLAY"),
          new BasicNameValuePair("zip-entries[1][_filename]", "layouts/badSchema.txt"),
          new BasicNameValuePair("zip-entries[1][_data][name]", "test-o")

      )));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().is2xxSuccessful());
      waitForProjectToFinishLoading(TEST_PROJECT);

      Project project = projectService.getProjectByProjectId(TEST_PROJECT);
      assertEquals(ProjectStatus.FAIL, project.getStatus());

    } finally {
      waitForProjectToFinishLoading(TEST_PROJECT);
    }
  }

  @Test
  public void testGetBackgrounds() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    createAndPersistProject(TEST_PROJECT);
    RequestBuilder request = get("/api/projects/{projectId}/backgrounds/", TEST_PROJECT)
        .session(session);
    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_backgrounds/get_all",
            projectPathParameters(),
            responseFields().andWithPrefix("[].", getBackgroundFields())));
  }

  private List<FieldDescriptor> getUpdateBackgroundFields() {
    return Arrays.asList(
        fieldWithPath("id")
            .type(JsonFieldType.NUMBER)
            .description("identifier"),
        fieldWithPath("name")
            .type(JsonFieldType.STRING)
            .description("name"),
        fieldWithPath("description")
            .type(JsonFieldType.STRING)
            .description("description"),
        fieldWithPath("creator")
            .type(JsonFieldType.STRING)
            .description("who created background"),
        fieldWithPath("order")
            .type(JsonFieldType.NUMBER)
            .description("order used when listing all backgrounds"),
        fieldWithPath("defaultOverlay")
            .type(JsonFieldType.BOOLEAN)
            .description(
                "should the background be used as default (at most one per project should be marked with true)"));
  }

  private List<FieldDescriptor> getBackgroundFields() {
    return Arrays.asList(
        fieldWithPath("id")
            .type(JsonFieldType.NUMBER)
            .description("identifier"),
        fieldWithPath("name")
            .type(JsonFieldType.STRING)
            .description("name"),
        fieldWithPath("description")
            .type(JsonFieldType.STRING)
            .optional()
            .description("description"),
        fieldWithPath("project.projectId")
            .type(JsonFieldType.STRING)
            .description("project id where this background belongs to"),
        fieldWithPath("creator.login")
            .type(JsonFieldType.STRING)
            .description("who created background"),
        fieldWithPath("status")
            .type(JsonFieldType.STRING)
            .description("is the background ready. Available statuses are: "
                + snippets.getOptionsAsString(ProjectBackgroundStatus.class)),
        fieldWithPath("progress")
            .type(JsonFieldType.NUMBER)
            .description("generating images progress information (in %)"),
        fieldWithPath("order")
            .type(JsonFieldType.NUMBER)
            .description("order used when listing all backgrounds"),
        fieldWithPath("images[].id")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath("images[].projectBackground.id")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath("images[].path")
            .type(JsonFieldType.STRING)
            .description("directory where background tiles are located"),
        fieldWithPath("images[].model.id")
            .type(JsonFieldType.NUMBER)
            .description("(sub)map for which images are described"),
        fieldWithPath("defaultOverlay")
            .type(JsonFieldType.BOOLEAN)
            .description(
                "should the background be used as default (at most one per project should be marked with true)"));
  }

  @Test
  public void testGetBackgroundById() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    Project project = createAndPersistProject(TEST_PROJECT);
    RequestBuilder request = get("/api/projects/{projectId}/backgrounds/{backgroundId}", TEST_PROJECT,
        project.getProjectBackgrounds().get(0).getId())
            .session(session);
    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_backgrounds/get_by_id",
            backgroundPathParameters(),
            responseFields(getBackgroundFields())));
  }

  @Test
  public void testUpdateBackgroundById() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    Project project = createAndPersistProject(TEST_PROJECT);
    ProjectBackground background = new ProjectBackground("weird_title");
    background.setId(project.getProjectBackgrounds().get(0).getId());
    background.setCreator(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    background.setDescription("new description");

    ObjectMapper mapper = new ObjectMapper();
    mapper.addMixIn(ProjectBackground.class, ProjectBackgroundUpdateMixIn.class);

    RequestBuilder request = patch("/api/projects/{projectId}/backgrounds/{backgroundId}", TEST_PROJECT,
        project.getProjectBackgrounds().get(0).getId())
            .contentType(MediaType.APPLICATION_JSON)
            .session(session)
            .content(mapper.writeValueAsString(background));

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_backgrounds/update_background",
            backgroundPathParameters(),
            requestFields(getUpdateBackgroundFields()),
            responseFields(getBackgroundFields())))
        .andReturn().getResponse().getContentAsString();

    ProjectBackground newBackground = objectMapper.readValue(response, new TypeReference<ProjectBackground>() {
    });
    assertEquals(background.getName(), newBackground.getName());
  }

  @Test
  public void testRemoveBackgroundById() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        admin);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);
    mockMvc.perform(request).andExpect(status().is2xxSuccessful());

    waitForProjectToFinishLoading(TEST_PROJECT);

    Project project = projectService.getProjectByProjectId(TEST_PROJECT);
    List<ProjectBackground> backgrounds = projectBackgroundService.getProjectBackgroundsByProject(project);
    ProjectBackground background = backgrounds.get(0);
    String path = "src/main/webapp/../map_images/" + project.getDirectory() + "/"
        + background.getProjectBackgroundImageLayer().iterator().next().getDirectory();
    assertTrue(new File(path).exists());

    request = delete("/api/projects/{projectId}/backgrounds/{backgroundId}", TEST_PROJECT,
        backgrounds.get(0).getId())
            .session(session);
    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_backgrounds/delete_background",
            backgroundPathParameters(),
            responseFields()));
    assertFalse(new File(path).exists());

    int backgroundCount = projectBackgroundService.getProjectBackgroundsByProject(project).size();
    assertEquals(backgrounds.size() - 1, backgroundCount);
  }

  private PathParametersSnippet backgroundPathParameters() {
    return projectPathParameters().and(parameterWithName("backgroundId").description("background identifier"));
  }

  private void waitForProjectToFinishRemoving(final String projectId) throws InterruptedException {
    while (projectService.projectExists(projectId)) {
      Thread.sleep(1);
    }
  }

  @Test(timeout = 10000)
  public void checkDefaultPrivilegesAfterAddingProject() throws Exception {
    User user = createCurator(USER_LOGIN, USER_PASSWORD);
    grantPrivilegeInSeparateThread(user, PrivilegeType.READ_PROJECT.name() + ":*");
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        curator);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(CURATOR_LOGIN, CURATOR_PASSWORD));

    mockMvc.perform(request).andExpect(status().is2xxSuccessful());

    waitForProjectToFinishLoading(TEST_PROJECT);

    User u = userService.getUserByLogin(USER_LOGIN, true);
    User c = userService.getUserByLogin(CURATOR_LOGIN, true);
    assertTrue("User privileges weren't updated after project was created",
        u.getPrivileges().contains(new Privilege(PrivilegeType.READ_PROJECT, TEST_PROJECT)));
    assertTrue("Curator privileges weren't updated after project was created",
        c.getPrivileges().contains(new Privilege(PrivilegeType.WRITE_PROJECT, TEST_PROJECT)));
    assertTrue("Curator privileges weren't updated after project was created",
        c.getPrivileges().contains(new Privilege(PrivilegeType.READ_PROJECT, TEST_PROJECT)));
  }

  @Test(timeout = 10000)
  public void checkDropPrivilegesAfterRemovingProject() throws Exception {
    createAndPersistProject(TEST_PROJECT);
    grantPrivilegeInSeparateThread(curator, PrivilegeType.WRITE_PROJECT.name() + ":" + TEST_PROJECT);
    RequestBuilder request = delete("/api/projects/" + TEST_PROJECT)
        .session(createSession(CURATOR_LOGIN, CURATOR_PASSWORD));

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    waitForProjectToFinishRemoving(TEST_PROJECT);

    User c = userService.getUserByLogin(CURATOR_LOGIN, true);

    assertFalse("Curator privileges weren't updated after project was removed",
        c.getPrivileges().contains(new Privilege(PrivilegeType.WRITE_PROJECT, TEST_PROJECT)));
  }

  private void grantPrivilegeInSeparateThread(final User user, final String string) throws Exception {
    userService.grantUserPrivilege(user, PrivilegeType.valueOf(string.split(":")[0]), string.split(":")[1]);
  }

  @Test(timeout = 10000)
  public void addProjectWithInvalidId() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        admin);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/*")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

    mockMvc.perform(request).andExpect(status().is4xxClientError());
    assertNull(projectService.getProjectByProjectId("*"));
  }

  @Test(timeout = 10000)
  public void addProjectWithoutPrivileges() throws Exception {
    User user = createUser(USER_LOGIN, USER_PASSWORD);
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        user);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(USER_LOGIN, USER_PASSWORD));

    mockMvc.perform(request).andExpect(status().isForbidden());
    assertNull(projectService.getProjectByProjectId(TEST_PROJECT));
  }

  @Test
  public void testUserPrivilegeChangeDuringUpload() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);

    Project defaultProject = createAndPersistProject(TEST_PROJECT_2);
    grantPrivilegeInSeparateThread(admin, PrivilegeType.READ_PROJECT.name() + ":" + defaultProject.getProjectId());
    grantPrivilegeInSeparateThread(admin, PrivilegeType.WRITE_PROJECT.name() + ":" + defaultProject.getProjectId());

    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        curator);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser",
            "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(CURATOR_LOGIN, CURATOR_PASSWORD));

    mockMvc.perform(request).andExpect(status().is2xxSuccessful());
    waitForProjectToFinishLoading(TEST_PROJECT);

    userService.revokeObjectDomainPrivilegesForAllUsers(PrivilegeType.READ_PROJECT,
        defaultProject.getProjectId());

    Project project = projectService.getProjectByProjectId(TEST_PROJECT);
    assertEquals(ProjectStatus.DONE, project.getStatus());
  }

  @Test
  public void createWithoutFileId() throws Exception {
    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("projectId", TEST_PROJECT),
        new BasicNameValuePair("name", "New Disease Map"),
        new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
        new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

    RequestBuilder request = post("/api/projects/{projectId}/", TEST_PROJECT)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

    mockMvc.perform(request).andExpect(status().isBadRequest());
  }

  @Test
  public void createWithoNotOwnFileFileId() throws Exception {
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        curator);

    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("projectId", TEST_PROJECT),
          new BasicNameValuePair("name", "New Disease Map"),
          new BasicNameValuePair("file-id", fileEntry.getId() + ""),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

      RequestBuilder request = post("/api/projects/{projectId}/", TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().isForbidden());
    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void createWithoNoParser() throws Exception {
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        curator);

    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("projectId", TEST_PROJECT),
          new BasicNameValuePair("name", "New Disease Map"),
          new BasicNameValuePair("file-id", fileEntry.getId() + ""),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"))));

      RequestBuilder request = post("/api/projects/{projectId}/", TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(CURATOR_LOGIN, CURATOR_PASSWORD));

      mockMvc.perform(request).andExpect(status().isBadRequest());
    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void createWithInvalidProjectId() throws Exception {
    UploadedFileEntry fileEntry = createFile(
        new String(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")), "UTF-8"),
        curator);

    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("projectId", "bla bla"),
          new BasicNameValuePair("name", "New Disease Map"),
          new BasicNameValuePair("file-id", fileEntry.getId() + ""),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"))));

      RequestBuilder request = post("/api/projects/{projectId}/", "bla bla")
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(createSession(CURATOR_LOGIN, CURATOR_PASSWORD));

      mockMvc.perform(request).andExpect(status().isBadRequest());
    } finally {
      removeFile(fileEntry);
    }
  }

  @Test
  public void testGetBackgroundByIdWithOtherUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    Project project = createAndPersistProject(TEST_PROJECT);
    ProjectBackground background = project.getProjectBackgrounds().get(0);
    background.setCreator(curator);
    projectBackgroundService.updateProjectBackground(background);

    RequestBuilder request = get("/api/projects/{projectId}/backgrounds/{backgroundId}",
        TEST_PROJECT,
        background.getId())
            .session(session);
    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void addComplexProjectWithGenomicLayout() throws Exception {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    UploadedFileEntry fileEntry = createFile(
        Files.readAllBytes(Paths.get("./src/test/resources/complex_model_with_genomic_layout.zip")), admin);
    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("file-id", String.valueOf(fileEntry.getId())),
          new BasicNameValuePair("mapCanvasType", "OPEN_LAYERS"),
          new BasicNameValuePair("parser", "lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser"),

          new BasicNameValuePair("zip-entries[0][_type]", "MAP"),
          new BasicNameValuePair("zip-entries[0][_filename]", "main.xml"),
          new BasicNameValuePair("zip-entries[0][_data][root]", "true"),
          new BasicNameValuePair("zip-entries[0][_data][name]", "s1"),
          new BasicNameValuePair("zip-entries[0][_data][type][id]", "UNKNOWN"),
          new BasicNameValuePair("zip-entries[0][_data][type][name]", "Unknown"),

          new BasicNameValuePair("zip-entries[1][_type]", "OVERLAY"),
          new BasicNameValuePair("zip-entries[1][_filename]", "layouts/goodSchema.txt"),
          new BasicNameValuePair("zip-entries[1][_data][name]", "test-o")

      )));

      RequestBuilder request = post("/api/projects/" + TEST_PROJECT)
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(session);

      mockMvc.perform(request).andExpect(status().is2xxSuccessful());
      waitForProjectToFinishLoading(TEST_PROJECT);

      RequestBuilder overlayRequest = get("/api/projects/{projectId}/overlays/", TEST_PROJECT)
          .session(session);

      String json = mockMvc.perform(overlayRequest)
          .andExpect(status().is2xxSuccessful())
          .andReturn().getResponse().getContentAsString();
      List<Map<String, Object>> overlays = objectMapper.readValue(json, new TypeReference<List<Map<String, Object>>>() {
      });

      logger.debug(json);

      assertNotNull(overlays.get(0).get("genomeType"));
      assertNotNull(overlays.get(0).get("genomeVersion"));
    } finally {
      waitForProjectToFinishLoading(TEST_PROJECT);
    }
  }

}
